# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os


class Creator:
    def __init__(
        self,
        path,
        config=None,
        cache=None,
        print=print,
        statistics_tmp=None,
        overwrite=False,
        **kwargs,
    ):
        self.path = path  # Output path
        self.config = config
        self.cache = cache
        self.print = print
        self.statistics_tmp = statistics_tmp
        self.overwrite = overwrite

    def init(self, check_name=False):
        # check path
        _, ext = os.path.splitext(self.path)
        assert ext != "zarr", f"Unsupported extension={ext}"
        from .loaders import InitialiseLoader

        if self._path_readable() and not self.overwrite:
            raise Exception(f"{self.path} already exists. Use overwrite=True to overwrite.")

        with self._cache_context():
            obj = InitialiseLoader.from_config(
                path=self.path,
                config=self.config,
                statistics_tmp=self.statistics_tmp,
                print=self.print,
            )
            obj.initialise(check_name=check_name)

    def load(self, parts=None):
        from .loaders import ContentLoader

        with self._cache_context():
            loader = ContentLoader.from_dataset_config(
                path=self.path,
                statistics_tmp=self.statistics_tmp,
                print=self.print,
                parts=parts,
            )
            loader.load()

    def statistics(self, force=False, output=None, start=None, end=None):
        from .loaders import StatisticsLoader

        loader = StatisticsLoader.from_dataset(
            path=self.path,
            print=self.print,
            force=force,
            statistics_tmp=self.statistics_tmp,
            statistics_output=output,
            recompute=False,
            statistics_start=start,
            statistics_end=end,
        )
        loader.run()

    def size(self):
        from .loaders import SizeLoader

        loader = SizeLoader.from_dataset(path=self.path, print=self.print)
        loader.add_total_size()

    def cleanup(self):
        from .loaders import CleanupLoader

        loader = CleanupLoader.from_dataset(
            path=self.path,
            print=self.print,
            statistics_tmp=self.statistics_tmp,
        )
        loader.run()

    def patch(self, **kwargs):
        from .patch import apply_patch

        apply_patch(self.path, **kwargs)

    def finalise(self, **kwargs):
        self.statistics(**kwargs)
        self.size()

    def create(self):
        self.init()
        self.load()
        self.finalise()
        self.cleanup()

    def _cache_context(self):
        from .utils import cache_context

        return cache_context(self.cache)

    def _path_readable(self):
        import zarr

        try:
            zarr.open(self.path, "r")
            return True
        except zarr.errors.PathNotFoundError:
            return False



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

import tqdm

from . import Command

LOG = logging.getLogger(__name__)

try:
    isatty = sys.stdout.isatty() and os.environ.get("TERM") != "dumb"
except AttributeError:
    isatty = False

"""

~/.aws/credentials

[default]
endpoint_url = https://object-store.os-api.cci1.ecmwf.int
aws_access_key_id=xxx
aws_secret_access_key=xxxx

Then:

anemoi-datasets copy aifs-ea-an-oper-0001-mars-o96-1979-2022-1h-v3.zarr/
    s3://ml-datasets/stable/aifs-ea-an-oper-0001-mars-o96-1979-2022-1h-v3.zarr

zinfo https://object-store.os-api.cci1.ecmwf.int/
    ml-datasets/stable/aifs-ea-an-oper-0001-mars-o96-1979-2022-1h-v3.zarr
"""


class CopyMixin:
    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument("--transfers", type=int, default=8)
        command_parser.add_argument("--block-size", type=int, default=100)
        command_parser.add_argument("--overwrite", action="store_true")
        command_parser.add_argument("--progress", action="store_true")
        command_parser.add_argument("--nested", action="store_true", help="Use ZARR's nested directpry backend.")
        command_parser.add_argument(
            "--rechunk",
            nargs="+",
            help="Rechunk given array.",
            metavar="array=i,j,k,l",
        )
        command_parser.add_argument("source")
        command_parser.add_argument("target")

    def _store(self, path, nested=False):
        if nested:
            import zarr

            return zarr.storage.NestedDirectoryStore(path)
        return path

    def copy_chunk(self, n, m, source, target, block_size, _copy, progress):
        if _copy[n:m].all():
            LOG.info(f"Skipping {n} to {m}")
            return None

        for i in tqdm.tqdm(
            range(n, m),
            desc=f"Copying {n} to {m}",
            leave=False,
            disable=not isatty and not progress,
        ):
            target[i] = source[i]
        return slice(n, m)

    def copy_data(self, source, target, transfers, block_size, _copy, progress, rechunking):
        LOG.info("Copying data")
        source_data = source["data"]

        chunks = list(source_data.chunks)
        if "data" in rechunking:
            assert len(chunks) == len(rechunking["data"]), (chunks, rechunking["data"])
            for i, c in enumerate(rechunking["data"]):
                if c != -1:
                    chunks[i] = c

        target_data = (
            target["data"]
            if "data" in target
            else target.create_dataset(
                "data",
                shape=source_data.shape,
                chunks=chunks,
                dtype=source_data.dtype,
            )
        )

        executor = ThreadPoolExecutor(max_workers=transfers)
        tasks = []
        n = 0
        while n < target_data.shape[0]:
            tasks.append(
                executor.submit(
                    self.copy_chunk,
                    n,
                    min(n + block_size, target_data.shape[0]),
                    source_data,
                    target_data,
                    block_size,
                    _copy,
                    progress,
                )
            )
            n += block_size

        for future in tqdm.tqdm(as_completed(tasks), total=len(tasks), smoothing=0):
            copied = future.result()
            if copied is not None:
                _copy[copied] = True
                target["_copy"] = _copy

        target["_copy"] = _copy

        LOG.info("Copied data")

    def copy_array(self, name, source, target, transfers, block_size, _copy, progress, rechunking):
        for k, v in source.attrs.items():
            target.attrs[k] = v

        if name == "_copy":
            return

        if name == "data":
            self.copy_data(source, target, transfers, block_size, _copy, progress, rechunking)
            return

        LOG.info(f"Copying {name}")
        target[name] = source[name]
        LOG.info(f"Copied {name}")

    def copy_group(self, source, target, transfers, block_size, _copy, progress, rechunking):
        import zarr

        for k, v in source.attrs.items():
            target.attrs[k] = v

        for name in sorted(source.keys()):
            if isinstance(source[name], zarr.hierarchy.Group):
                group = target[name] if name in target else target.create_group(name)
                self.copy_group(
                    source[name],
                    group,
                    transfers,
                    block_size,
                    _copy,
                    progress,
                    rechunking,
                )
            else:
                self.copy_array(
                    name,
                    source,
                    target,
                    transfers,
                    block_size,
                    _copy,
                    progress,
                    rechunking,
                )

    def copy(self, source, target, transfers, block_size, progress, rechunking):
        import zarr

        if "_copy" not in target:
            target["_copy"] = zarr.zeros(
                source["data"].shape[0],
                dtype=bool,
            )
        _copy = target["_copy"]
        _copy_np = _copy[:]

        self.copy_group(source, target, transfers, block_size, _copy_np, progress, rechunking)
        del target["_copy"]

    def run(self, args):
        import zarr

        # base, ext = os.path.splitext(os.path.basename(args.source))
        # assert ext == ".zarr", ext
        # assert "." not in base, base
        LOG.info(f"Copying {args.source} to {args.target}")

        rechunking = {}
        if args.rechunk:
            for r in args.rechunk:
                k, v = r.split("=")
                if k != "data":
                    raise ValueError(f"Only rechunking data is supported: {k}")
                values = v.split(",")
                values = [-1 if x == "" else x for x in values]
                values = tuple(int(x) for x in values)
                rechunking[k] = values
            for k, v in rechunking.items():
                LOG.info(f"Rechunking {k} to {v}")

        try:
            target = zarr.open(self._store(args.target, args.nested), mode="r")
            if "_copy" in target:
                done = sum(1 if x else 0 for x in target["_copy"])
                todo = len(target["_copy"])
                LOG.info(
                    "Resuming copy, done %s out or %s, %s%%",
                    done,
                    todo,
                    int(done / todo * 100 + 0.5),
                )
            elif "sums" in target and "data" in target:  # sums is copied last
                LOG.error("Target already exists")
                return
        except ValueError as e:
            LOG.info(f"Target does not exist: {e}")
            pass

        source = zarr.open(self._store(args.source), mode="r")
        if args.overwrite:
            target = zarr.open(self._store(args.target, args.nested), mode="w")
        else:
            try:
                target = zarr.open(self._store(args.target, args.nested), mode="w+")
            except ValueError:
                target = zarr.open(self._store(args.target, args.nested), mode="w")
        self.copy(source, target, args.transfers, args.block_size, args.progress, rechunking)


class Copy(CopyMixin, Command):
    pass


command = Copy



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict

import numpy as np
from climetlab.indexing.fieldset import FieldArray


def normalise(x):
    return max(min(x, 1.0), -1.0)


def normalise_longitude(lon, minimum):
    while lon < minimum:
        lon += 360

    while lon >= minimum + 360:
        lon -= 360

    return lon


def rotate_winds(
    lats,
    lons,
    raw_lats,
    raw_lons,
    x_wind,
    y_wind,
    south_pole_latitude,
    south_pole_longitude,
    south_pole_rotation_angle=0,
):
    # Code from MIR
    assert south_pole_rotation_angle == 0
    C = np.deg2rad(90 - south_pole_latitude)
    cos_C = np.cos(C)
    sin_C = np.sin(C)

    new_x = np.zeros_like(x_wind)
    new_y = np.zeros_like(y_wind)

    for i, (vx, vy, lat, lon, raw_lat, raw_lon) in enumerate(zip(x_wind, y_wind, lats, lons, raw_lats, raw_lons)):
        lonRotated = south_pole_longitude - lon
        lon_rotated = normalise_longitude(lonRotated, -180)
        lon_unrotated = raw_lon

        a = np.deg2rad(lon_rotated)
        b = np.deg2rad(lon_unrotated)
        q = 1 if (sin_C * lon_rotated < 0.0) else -1.0  # correct quadrant

        cos_c = normalise(np.cos(a) * np.cos(b) + np.sin(a) * np.sin(b) * cos_C)
        sin_c = q * np.sqrt(1.0 - cos_c * cos_c)

        new_x[i] = cos_c * vx + sin_c * vy
        new_y[i] = -sin_c * vx + cos_c * vy

    return new_x, new_y


class NewDataField:
    def __init__(self, field, data):
        self.field = field
        self.data = data

    def to_numpy(self, *args, **kwargs):
        return self.data

    def __getattr__(self, name):
        return getattr(self.field, name)


def execute(context, input, u, v):
    """
    Unrotate the wind components of a GRIB file.
    """
    result = FieldArray()

    wind_params = (u, v)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.as_mars()
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    for _, pairs in wind_pairs.items():
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        x = pairs[u]
        y = pairs[v]

        lats, lons = x.grid_points()
        raw_lats, raw_longs = x.grid_points_raw()

        assert x.rotation == y.rotation

        u_new, v_new = rotate_winds(
            lats,
            lons,
            raw_lats,
            raw_longs,
            x.to_numpy(reshape=False),
            y.to_numpy(reshape=False),
            *x.rotation,
        )

        result.append(NewDataField(x, u_new))
        result.append(NewDataField(y, v_new))

    return result


if __name__ == "__main__":
    from climetlab import load_source

    source = load_source(
        "mars",
        date=-1,
        param="10u/10v",
        levtype="sfc",
        grid=[1, 1],
        area=[28, 0, -14, 40],
        rotation=[-22, -40],
    )

    source.save("source.grib")

    execute(None, source, "10u", "10v")



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict

from climetlab.indexing.fieldset import FieldArray


def rotate_winds(lats, lons, x_wind, y_wind, source_projection, target_projection):
    """
    Code provided by MetNO
    """
    import numpy as np
    import pyproj

    if source_projection == target_projection:
        return x_wind, x_wind

    source_projection = pyproj.Proj(source_projection)
    target_projection = pyproj.Proj(target_projection)

    transformer = pyproj.transformer.Transformer.from_proj(source_projection, target_projection)

    # To compute the new vector components:
    # 1) perturb each position in the direction of the winds
    # 2) convert the perturbed positions into the new coordinate system
    # 3) measure the new x/y components.
    #
    # A complication occurs when using the longlat "projections", since this is not a cartesian grid
    # (i.e. distances in each direction is not consistent), we need to deal with the fact that the
    # width of a longitude varies with latitude
    orig_speed = np.sqrt(x_wind**2 + y_wind**2)

    x0, y0 = source_projection(lons, lats)

    if source_projection.name != "longlat":
        x1 = x0 + x_wind
        y1 = y0 + y_wind
    else:
        # Reduce the perturbation, since x_wind and y_wind are in meters, which would create
        # large perturbations in lat, lon. Also, deal with the fact that the width of longitude
        # varies with latitude.
        factor = 3600000.0
        x1 = x0 + x_wind / factor / np.cos(np.deg2rad(lats))
        y1 = y0 + y_wind / factor

    X0, Y0 = transformer.transform(x0, y0)
    X1, Y1 = transformer.transform(x1, y1)

    new_x_wind = X1 - X0
    new_y_wind = Y1 - Y0
    if target_projection.name == "longlat":
        new_x_wind *= np.cos(np.deg2rad(lats))

    if target_projection.name == "longlat" or source_projection.name == "longlat":
        # Ensure the wind speed is not changed (which might not the case since the units in longlat
        # is degrees, not meters)
        curr_speed = np.sqrt(new_x_wind**2 + new_y_wind**2)
        new_x_wind *= orig_speed / curr_speed
        new_y_wind *= orig_speed / curr_speed

    return new_x_wind, new_y_wind


class NewDataField:
    def __init__(self, field, data):
        self.field = field
        self.data = data

    def to_numpy(self, *args, **kwargs):
        return self.data

    def __getattr__(self, name):
        return getattr(self.field, name)


def execute(
    context,
    input,
    x_wind,
    y_wind,
    source_projection=None,
    target_projection="+proj=longlat",
):
    from pyproj import CRS

    result = FieldArray()

    wind_params = (x_wind, y_wind)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.as_mars()
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    for _, pairs in wind_pairs.items():
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        x = pairs[x_wind]
        y = pairs[y_wind]

        assert x.grid_mapping == y.grid_mapping

        lats, lons = x.grid_points()
        x_new, y_new = rotate_winds(
            lats,
            lons,
            x.to_numpy(reshape=False),
            y.to_numpy(reshape=False),
            (source_projection if source_projection is not None else CRS.from_cf(x.grid_mapping)),
            target_projection,
        )

        result.append(NewDataField(x, x_new))
        result.append(NewDataField(y, y_new))

    return result



#!/usr/bin/env python
# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from anemoi.datasets import open_dataset

from . import Command


class Compare(Command):
    def add_arguments(self, command_parser):
        command_parser.add_argument("dataset1")
        command_parser.add_argument("dataset2")

    def run(self, args):
        ds1 = open_dataset(args.dataset1)
        ds2 = open_dataset(args.dataset2)

        v1 = set(ds1.variables)
        v2 = set(ds2.variables)

        print("Only in dataset 1:", ", ".join(sorted(v1 - v2)))
        print("Only in dataset 2:", ", ".join(sorted(v2 - v1)))
        print()
        common = sorted(v1 & v2)
        print("Common:")
        print("-------")
        print()

        for v in common:
            print(
                f"{v:14}",
                f"{ds1.statistics['mean'][ds1.name_to_index[v]]:14g}",
                f"{ds2.statistics['mean'][ds2.name_to_index[v]]:14g}",
            )


command = Compare



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import re

from climetlab.indexing.fieldset import FieldArray


class RenamedFieldMapping:
    def __init__(self, field, what, renaming):
        self.field = field
        self.what = what
        self.renaming = renaming

    def metadata(self, key):
        value = self.field.metadata(key)
        if key == self.what:
            return self.renaming.get(value, value)
        return value

    def __getattr__(self, name):
        return getattr(self.field, name)


class RenamedFieldFormat:
    def __init__(self, field, format):
        self.field = field
        self.format = format
        self.bits = re.findall(r"{(\w+)}", format)

    def metadata(self, key):
        value = self.field.metadata(key)
        if "{" + key + "}" in self.format:
            bits = {b: self.field.metadata(b) for b in self.bits}
            return self.format.format(**bits)
        return value

    def __getattr__(self, name):
        return getattr(self.field, name)


def execute(context, input, what="param", **kwargs):
    # print('🍍🍍🍍🍍🍍🍍🍍🍍🍍🍍🍍🍍🍍 ==========', kwargs)
    if what in kwargs:
        return FieldArray([RenamedFieldFormat(fs, kwargs[what]) for fs in input])

    return FieldArray([RenamedFieldMapping(fs, what, kwargs) for fs in input])



import fnmatch
import os
import sys
from collections import defaultdict

import climetlab as cml
import tqdm
import yaml

from . import Command

KEYS = ("class", "type", "stream", "expver", "levtype", "domain")


class Scan(Command):
    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--match",
            help="Give a glob pattern to match files (default: *.grib)",
            default="*.grib",
        )
        command_parser.add_argument(
            "--what",
            default="grib",
        )
        command_parser.add_argument("paths", nargs="+", help="Paths to scan")

    def run(self, args):
        def match(path):
            return fnmatch.fnmatch(path, args.match)

        paths = []
        for path in args.paths:
            if os.path.isfile(path):
                paths.append(path)
            else:
                for root, _, files in os.walk(path):
                    for file in files:
                        full = os.path.join(root, file)
                        paths.append(full)

        dates = set()
        gribs = defaultdict(set)
        unique = defaultdict(lambda: defaultdict(set))
        what = args.what

        for path in tqdm.tqdm(paths, leave=False):
            if not match(path):
                continue
            for field in tqdm.tqdm(cml.load_source("file", path), leave=False):
                dates.add(field.valid_datetime())
                mars = field.as_mars()
                keys = tuple(mars.get(k) for k in KEYS)
                gribs[keys].add(path)
                for k, v in mars.items():
                    if k not in KEYS + ("date", "time", "step"):
                        unique[keys][k].add(v)

        config = dict(
            description=f"Generated by {sys.argv})",
            dataset_status="experimental",
            name="test",
            dates=dict(values=sorted(dates)),
            input=dict(join=[]),
            output=dict(
                chunking=dict(dates=1, ensembles=1),
                dtype="float32",
                flatten_grid=True,
                order_by=["valid_datetime", "param_level", "number"],
                statistics="param_level",
                statistics_end=2020,
                remapping=dict(param_level="{param}_{levelist}"),
            ),
        )

        for k, v in sorted(gribs.items()):
            request = {what: dict(path=sorted(v), **dict(zip(KEYS, k)))}
            for k, v in sorted(unique[k].items()):
                if len(v) == 1:
                    request[what][k] = list(v)[0]
                else:
                    request[what][k] = sorted(v)

            config["input"]["join"].append(request)

        with open("scan-config.yaml", "w") as f:
            print(yaml.dump(config, sort_keys=False), file=f)


command = Scan



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np


def plot_mask(path, mask, lats, lons, global_lats, global_lons):
    import matplotlib.pyplot as plt

    middle = (np.amin(lons) + np.amax(lons)) / 2
    print("middle", middle)
    s = 1

    # gmiddle = (np.amin(global_lons)+ np.amax(global_lons))/2

    # print('gmiddle', gmiddle)
    # global_lons = global_lons-gmiddle+middle
    global_lons[global_lons >= 180] -= 360

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons, global_lats, s=s, marker="o", c="r")
    plt.savefig(path + "-global.png")

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="k")
    plt.savefig(path + "-cutout.png")

    plt.figure(figsize=(10, 5))
    plt.scatter(lons, lats, s=s)
    plt.savefig(path + "-lam.png")
    # plt.scatter(lons, lats, s=0.01)

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="r")
    plt.scatter(lons, lats, s=s)
    plt.savefig(path + "-both.png")
    # plt.scatter(lons, lats, s=0.01)


def latlon_to_xyz(lat, lon, radius=1.0):
    # https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
    # We assume that the Earth is a sphere of radius 1 so N(phi) = 1
    # We assume h = 0
    #
    phi = np.deg2rad(lat)
    lda = np.deg2rad(lon)

    cos_phi = np.cos(phi)
    cos_lda = np.cos(lda)
    sin_phi = np.sin(phi)
    sin_lda = np.sin(lda)

    x = cos_phi * cos_lda * radius
    y = cos_phi * sin_lda * radius
    z = sin_phi * radius

    return x, y, z


class Triangle3D:
    def __init__(self, v0, v1, v2):
        self.v0 = v0
        self.v1 = v1
        self.v2 = v2

    def intersect(self, ray_origin, ray_direction):
        # Möller–Trumbore intersection algorithm
        # https://en.wikipedia.org/wiki/M%C3%B6ller%E2%80%93Trumbore_intersection_algorithm

        epsilon = 0.0000001

        h = np.cross(ray_direction, self.v2 - self.v0)
        a = np.dot(self.v1 - self.v0, h)

        if -epsilon < a < epsilon:
            return False

        f = 1.0 / a
        s = ray_origin - self.v0
        u = f * np.dot(s, h)

        if u < 0.0 or u > 1.0:
            return False

        q = np.cross(s, self.v1 - self.v0)
        v = f * np.dot(ray_direction, q)

        if v < 0.0 or u + v > 1.0:
            return False

        t = f * np.dot(self.v2 - self.v0, q)

        if t > epsilon:
            return True

        return False


def cropping_mask(lats, lons, north, west, south, east):
    mask = (
        (lats >= south)
        & (lats <= north)
        & (
            ((lons >= west) & (lons <= east))
            | ((lons >= west + 360) & (lons <= east + 360))
            | ((lons >= west - 360) & (lons <= east - 360))
        )
    )
    return mask


def cutout_mask(
    lats,
    lons,
    global_lats,
    global_lons,
    cropping_distance=2.0,
    min_distance_km=None,
    plot=None,
):
    """
    Return a mask for the points in [global_lats, global_lons] that are inside of [lats, lons]
    """
    from scipy.spatial import KDTree

    # TODO: transform min_distance from lat/lon to xyz

    assert global_lats.ndim == 1
    assert global_lons.ndim == 1
    assert lats.ndim == 1
    assert lons.ndim == 1

    assert global_lats.shape == global_lons.shape
    assert lats.shape == lons.shape

    north = np.amax(lats)
    south = np.amin(lats)
    east = np.amax(lons)
    west = np.amin(lons)

    # Reduce the global grid to the area of interest

    mask = cropping_mask(
        global_lats,
        global_lons,
        np.min([90.0, north + cropping_distance]),
        west - cropping_distance,
        np.max([-90.0, south - cropping_distance]),
        east + cropping_distance,
    )

    # return mask
    # mask = np.array([True] * len(global_lats), dtype=bool)
    global_lats_masked = global_lats[mask]
    global_lons_masked = global_lons[mask]

    global_xyx = latlon_to_xyz(global_lats_masked, global_lons_masked)
    global_points = np.array(global_xyx).transpose()

    xyx = latlon_to_xyz(lats, lons)
    lam_points = np.array(xyx).transpose()

    # Use a KDTree to find the nearest points
    kdtree = KDTree(lam_points)
    distances, indices = kdtree.query(global_points, k=3)

    if min_distance_km is not None:
        min_distance = min_distance_km / 6371.0
    else:
        # Estimnation of the minimum distance between two grib points

        glats = sorted(set(global_lats_masked))
        glons = sorted(set(global_lons_masked))
        min_dlats = np.min(np.diff(glats))
        min_dlons = np.min(np.diff(glons))

        # Use the center of the LAM grid as the reference point
        centre = np.mean(lats), np.mean(lons)
        centre_xyz = np.array(latlon_to_xyz(*centre))

        pt1 = np.array(latlon_to_xyz(centre[0] + min_dlats, centre[1]))
        pt2 = np.array(latlon_to_xyz(centre[0], centre[1] + min_dlons))
        min_distance = (
            min(
                np.linalg.norm(pt1 - centre_xyz),
                np.linalg.norm(pt2 - centre_xyz),
            )
            / 2.0
        )

    zero = np.array([0.0, 0.0, 0.0])
    ok = []
    for i, (global_point, distance, index) in enumerate(zip(global_points, distances, indices)):
        t = Triangle3D(lam_points[index[0]], lam_points[index[1]], lam_points[index[2]])
        # distance = np.min(distance)
        # The point is inside the triangle if the intersection with the ray
        # from the point to the center of the Earth is not None
        # (the direction of the ray is not important)

        intersect = t.intersect(zero, global_point)
        close = np.min(distance) <= min_distance

        ok.append(intersect or close)

    j = 0
    ok = np.array(ok)
    for i, m in enumerate(mask):
        if not m:
            continue

        mask[i] = ok[j]
        j += 1

    assert j == len(ok)

    # Invert the mask, so we have only the points outside the cutout
    mask = ~mask

    if plot:
        plot_mask(plot, mask, lats, lons, global_lats, global_lons)

    return mask


def thinning_mask(
    lats,
    lons,
    global_lats,
    global_lons,
    cropping_distance=2.0,
):
    """
    Return the list of points in [lats, lons] closest to [global_lats, global_lons]
    """
    from scipy.spatial import KDTree

    assert global_lats.ndim == 1
    assert global_lons.ndim == 1
    assert lats.ndim == 1
    assert lons.ndim == 1

    assert global_lats.shape == global_lons.shape
    assert lats.shape == lons.shape

    north = np.amax(lats)
    south = np.amin(lats)
    east = np.amax(lons)
    west = np.amin(lons)

    # Reduce the global grid to the area of interest

    mask = cropping_mask(
        global_lats,
        global_lons,
        np.min([90.0, north + cropping_distance]),
        west - cropping_distance,
        np.max([-90.0, south - cropping_distance]),
        east + cropping_distance,
    )

    # return mask
    global_lats_masked = global_lats[mask]
    global_lons_masked = global_lons[mask]

    global_xyx = latlon_to_xyz(global_lats_masked, global_lons_masked)
    global_points = np.array(global_xyx).transpose()

    xyx = latlon_to_xyz(lats, lons)
    points = np.array(xyx).transpose()

    # Use a KDTree to find the nearest points
    kdtree = KDTree(points)
    _, indices = kdtree.query(global_points, k=1)

    return np.array([i for i in indices])


if __name__ == "__main__":
    global_lats, global_lons = np.meshgrid(
        np.linspace(90, -90, 90),
        np.linspace(-180, 180, 180),
    )
    global_lats = global_lats.flatten()
    global_lons = global_lons.flatten()

    lats, lons = np.meshgrid(
        np.linspace(50, 40, 100),
        np.linspace(-10, 15, 100),
    )
    lats = lats.flatten()
    lons = lons.flatten()

    mask = cutout_mask(lats, lons, global_lats, global_lons, cropping_distance=5.0)

    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(10, 5))
    plt.scatter(global_lons, global_lats, s=0.01, marker="o", c="r")
    plt.scatter(global_lons[mask], global_lats[mask], s=0.1, c="k")
    # plt.scatter(lons, lats, s=0.01)
    plt.savefig("cutout.png")



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import logging
import os
from copy import deepcopy

import yaml
from climetlab.core.order import normalize_order_by

from .utils import load_json_or_yaml

LOG = logging.getLogger(__name__)


def _get_first_key_if_dict(x):
    if isinstance(x, str):
        return x
    return list(x.keys())[0]


def ensure_element_in_list(lst, elt, index):
    if elt in lst:
        assert lst[index] == elt
        return lst

    _lst = [_get_first_key_if_dict(d) for d in lst]
    if elt in _lst:
        assert _lst[index] == elt
        return lst

    return lst[:index] + [elt] + lst[index:]


def check_dict_value_and_set(dic, key, value):
    if key in dic:
        if dic[key] == value:
            return
        raise ValueError(f"Cannot use {key}={dic[key]}. Must use {value}.")
    print(f"Setting {key}={value} in config")
    dic[key] = value


class DictObj(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = DictObj(value)
                continue
            if isinstance(value, list):
                self[key] = [DictObj(item) if isinstance(item, dict) else item for item in value]
                continue

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self[attr] = value


def resolve_includes(config):
    if isinstance(config, list):
        return [resolve_includes(c) for c in config]
    if isinstance(config, dict):
        include = config.pop("<<", {})
        new = deepcopy(include)
        new.update(config)
        return {k: resolve_includes(v) for k, v in new.items()}
    return config


class Config(DictObj):
    def __init__(self, config=None, **kwargs):
        if isinstance(config, str):
            self.config_path = os.path.realpath(config)
            config = load_json_or_yaml(config)
        else:
            config = deepcopy(config if config is not None else {})
        config = resolve_includes(config)
        config.update(kwargs)
        super().__init__(config)


class OutputSpecs:
    def __init__(self, config, parent):
        self.config = config
        if "order_by" in config:
            assert isinstance(config.order_by, dict), config.order_by

        self.parent = parent

    @property
    def dtype(self):
        return self.config.dtype

    @property
    def order_by_as_list(self):
        # this is used when an ordered dict is not supported (e.g. zarr attributes)
        return [{k: v} for k, v in self.config.order_by.items()]

    def get_chunking(self, coords):
        user = deepcopy(self.config.chunking)
        chunks = []
        for k, v in coords.items():
            if k in user:
                chunks.append(user.pop(k))
            else:
                chunks.append(len(v))
        if user:
            raise ValueError(
                f"Unused chunking keys from config: {list(user.keys())}, not in known keys : {list(coords.keys())}"
            )
        return tuple(chunks)

    @property
    def order_by(self):
        return self.config.order_by

    @property
    def remapping(self):
        return self.config.remapping

    @property
    def flatten_grid(self):
        return self.config.flatten_grid

    @property
    def statistics(self):
        return self.config.statistics


class LoadersConfig(Config):
    def __init__(self, config, *args, **kwargs):

        super().__init__(config, *args, **kwargs)

        # TODO: should use a json schema to validate the config

        self.setdefault("dataset_status", "experimental")
        self.setdefault("description", "No description provided.")
        self.setdefault("licence", "unknown")
        self.setdefault("attribution", "unknown")

        self.setdefault("build", Config())
        self.build.setdefault("group_by", "monthly")

        self.setdefault("output", Config())
        self.output.setdefault("order_by", ["valid_datetime", "param_level", "number"])
        self.output.setdefault("remapping", Config(param_level="{param}_{levelist}"))
        self.output.setdefault("statistics", "param_level")
        self.output.setdefault("chunking", Config(dates=1, ensembles=1))
        self.output.setdefault("dtype", "float32")

        if "statistics_start" in self.output:
            raise ValueError("statistics_start is not supported anymore. Use 'statistics:start:' instead")
        if "statistics_end" in self.output:
            raise ValueError("statistics_end is not supported anymore. Use 'statistics:end:' instead")

        self.setdefault("statistics", Config())

        check_dict_value_and_set(self.output, "flatten_grid", True)
        check_dict_value_and_set(self.output, "ensemble_dimension", 2)

        assert isinstance(self.output.order_by, (list, tuple)), self.output.order_by
        self.output.order_by = ensure_element_in_list(self.output.order_by, "number", self.output.ensemble_dimension)

        order_by = self.output.order_by
        assert len(order_by) == 3, order_by
        assert _get_first_key_if_dict(order_by[0]) == "valid_datetime", order_by
        assert _get_first_key_if_dict(order_by[2]) == "number", order_by

        self.output.order_by = normalize_order_by(self.output.order_by)

        self.dates["group_by"] = self.build.group_by

        ###########

        self.reading_chunks = self.get("reading_chunks")

    def get_serialisable_dict(self):
        return _prepare_serialisation(self)


def _prepare_serialisation(o):
    if isinstance(o, dict):
        dic = {}
        for k, v in o.items():
            v = _prepare_serialisation(v)
            if k == "order_by" and isinstance(v, dict):
                # zarr attributes are saved with sort_keys=True
                # and ordered dict are reordered.
                # This is a problem for "order_by"
                # We ensure here that the order_by key contains
                # a list of dict
                v = [{kk: vv} for kk, vv in v.items()]
            dic[k] = v
        return dic

    if isinstance(o, (list, tuple)):
        return [_prepare_serialisation(v) for v in o]

    if o in (None, True, False):
        return o

    if isinstance(o, (str, int, float)):
        return o

    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()

    return str(o)


def loader_config(config):
    config = Config(config)
    obj = LoadersConfig(config)

    # yaml round trip to check that serialisation works as expected
    copy = obj.get_serialisable_dict()
    copy = yaml.load(yaml.dump(copy), Loader=yaml.SafeLoader)
    copy = Config(copy)
    copy = LoadersConfig(config)
    assert yaml.dump(obj) == yaml.dump(copy), (obj, copy)

    return copy


def build_output(*args, **kwargs):
    return OutputSpecs(*args, **kwargs)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from .debug import Node
from .debug import debug_indexing
from .forewards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class ConcatMixin:

    def __len__(self):
        return sum(len(i) for i in self.datasets)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        # print(index, changes)
        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(index[0], lengths)
        # print("slies", slices)
        result = [d[update_tuple(index, 0, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=0)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        # TODO: optimize
        k = 0
        while n >= self.datasets[k]._len:
            n -= self.datasets[k]._len
            k += 1
        return self.datasets[k][n]

    @debug_indexing
    def _get_slice(self, s):
        result = []

        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(s, lengths)

        result = [d[i] for (d, i) in zip(self.datasets, slices) if i is not None]

        return np.concatenate(result)

    @cached_property
    def missing(self):
        result = set()
        offset = 0
        for d in self.datasets:
            result = result | set(m + offset for m in d.missing)
            offset += len(d)
        return result


class Concat(ConcatMixin, Combined):

    def check_compatibility(self, d1, d2):
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def check_same_lengths(self, d1, d2):
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1, d2):
        # Turned off because we are concatenating along the dates axis
        pass

    @property
    def dates(self):
        return np.concatenate([d.dates for d in self.datasets])

    @property
    def shape(self):
        return (len(self),) + self.datasets[0].shape[1:]

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])


def concat_factory(args, kwargs, zarr_root):

    datasets = kwargs.pop("concat")
    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    assert isinstance(datasets, (list, tuple))

    datasets = [_open(e, zarr_root) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Concat(datasets)._subset(**kwargs)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab import load_source
from climetlab.utils.patterns import Pattern


def check(ds, paths, **kwargs):
    count = 1
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list)):
            count *= len(v)

    if len(ds) != count:
        raise ValueError(f"Expected {count} fields, got {len(ds)} (kwargs={kwargs}, paths={paths})")


def execute(context, dates, path, *args, **kwargs):
    given_paths = path if isinstance(path, list) else [path]

    ds = load_source("empty")
    dates = [d.isoformat() for d in dates]

    for path in given_paths:
        paths = Pattern(path, ignore_missing_keys=True).substitute(*args, date=dates, **kwargs)

        for name in ("grid", "area", "rotation", "frame", "resol", "bitmap"):
            if name in kwargs:
                raise ValueError(f"MARS interpolation parameter '{name}' not supported")

        for path in paths:
            context.trace("📁", "PATH", path)
            s = load_source("file", path)
            s = s.sel(valid_datetime=dates, **kwargs)
            ds = ds + s

    if kwargs:
        check(ds, given_paths, valid_datetime=dates, **kwargs)

    return ds



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import os
from copy import deepcopy
from functools import cached_property

import numpy as np
import semantic_version
import tqdm
from anemoi.utils.humanize import bytes
from anemoi.utils.humanize import number
from anemoi.utils.humanize import when
from anemoi.utils.text import dotted_line
from anemoi.utils.text import progress
from anemoi.utils.text import table

from anemoi.datasets import open_dataset
from anemoi.datasets.data.stores import open_zarr

LOG = logging.getLogger(__name__)


def compute_directory_size(path):
    if not os.path.isdir(path):
        return None, None
    size = 0
    n = 0
    for dirpath, _, filenames in tqdm.tqdm(os.walk(path), desc="Computing size", leave=False):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            size += os.path.getsize(file_path)
            n += 1
    return size, n


def local_time_bug(lon, date):
    delta = date - datetime.datetime(date.year, date.month, date.day)
    hours_since_midnight = delta.days + delta.seconds / 86400.0  # * 24 is missing
    return (lon / 360.0 * 24.0 + hours_since_midnight) % 24


def cos_local_time_bug(lon, date):
    radians = local_time_bug(lon, date) / 24 * np.pi * 2
    return np.cos(radians)


def find(config, name):
    if isinstance(config, dict):
        if name in config:
            return config[name]

        for k, v in config.items():
            r = find(v, name)
            if r is not None:
                return r

    if isinstance(config, list):
        for v in config:
            r = find(v, name)
            if r is not None:
                return r

    return None


class Version:
    def __init__(self, path, zarr, metadata, version):
        self.path = path
        self.zarr = zarr
        self.metadata = metadata
        self.version = version
        self.dataset = None
        # try:
        self.dataset = open_dataset(self.path)
        # except Exception as e:
        #     LOG.error("Error opening dataset '%s': %s", self.path, e)

    def describe(self):
        print(f"📦 Path          : {self.path}")
        print(f"🔢 Format version: {self.version}")

    def probe(self):
        if "cos_local_time" not in self.name_to_index:
            print("⚠️ probe: no cos_local_time")
            return

        try:
            lon = self.longitudes
        except AttributeError:
            print("⚠️ probe: no longitudes")
            return
        # print(json.dumps(self.metadata, indent=4))
        cos_local_time = self.name_to_index["cos_local_time"]
        data = self.data
        start, end, frequency = self.first_date, self.last_date, self.frequency
        date = start
        same = 0
        for i in range(10):
            field = data[i, cos_local_time]
            buggy = cos_local_time_bug(lon, date).reshape(field.hape)
            diff = np.abs(field - buggy)
            if np.max(diff) < 1e-5:
                same += 1
            date += datetime.timedelta(hours=frequency)
            if date > end:
                break
        if same > 1:
            print("❌ probe: cos_local_time is buggy")
            return

        print("✅ probe: cos_local_time is fixed")

    @property
    def name_to_index(self):
        return find(self.metadata, "name_to_index")

    @property
    def longitudes(self):
        try:
            return self.zarr.longitudes[:]
        except (KeyError, AttributeError):
            return self.zarr.longitude[:]

    @property
    def data(self):
        try:
            return self.zarr.data
        except AttributeError:
            return self.zarr

    @property
    def first_date(self):
        return datetime.datetime.fromisoformat(self.metadata["first_date"])

    @property
    def last_date(self):
        return datetime.datetime.fromisoformat(self.metadata["last_date"])

    @property
    def frequency(self):
        return self.metadata["frequency"]

    @property
    def resolution(self):
        return self.metadata["resolution"]

    @property
    def field_shape(self):
        return self.metadata.get("field_shape")

    @property
    def proj_string(self):
        return self.metadata.get("proj_string")

    @property
    def shape(self):
        if self.data and hasattr(self.data, "shape"):
            return self.data.shape

    @property
    def n_missing_dates(self):
        if "missing_dates" in self.metadata:
            return len(self.metadata["missing_dates"])
        return None

    @property
    def uncompressed_data_size(self):
        if self.data and hasattr(self.data, "dtype") and hasattr(self.data, "size"):
            return self.data.dtype.itemsize * self.data.size

    def info(self, detailed, size):
        print()
        print(f'📅 Start      : {self.first_date.strftime("%Y-%m-%d %H:%M")}')
        print(f'📅 End        : {self.last_date.strftime("%Y-%m-%d %H:%M")}')
        print(f"⏰ Frequency  : {self.frequency}h")
        if self.n_missing_dates is not None:
            print(f"🚫 Missing    : {self.n_missing_dates:,}")
        print(f"🌎 Resolution : {self.resolution}")
        print(f"🌎 Field shape: {self.field_shape}")

        print()
        shape_str = "📐 Shape      : "
        if self.shape:
            shape_str += " × ".join(["{:,}".format(s) for s in self.shape])
        if self.uncompressed_data_size:
            shape_str += f" ({bytes(self.uncompressed_data_size)})"
        print(shape_str)
        self.print_sizes(size)
        print()
        rows = []

        if self.statistics_ready:
            stats = self.statistics
        else:
            stats = [["-"] * len(self.variables)] * 4

        for i, v in enumerate(self.variables):
            rows.append([i, v] + [x[i] for x in stats])

        print(
            table(
                rows,
                header=["Index", "Variable", "Min", "Max", "Mean", "Stdev"],
                align=[">", "<", ">", ">", ">", ">"],
                margin=3,
            )
        )

        if detailed:
            self.details()

        self.progress()
        if self.ready():
            self.probe()

        print()

    @property
    def variables(self):
        return [v[0] for v in sorted(self.name_to_index.items(), key=lambda x: x[1])]

    @property
    def total_size(self):
        return self.zarr.attrs.get("total_size")

    @property
    def total_number_of_files(self):
        return self.zarr.attrs.get("total_number_of_files")

    def print_sizes(self, size):
        total_size = self.total_size
        n = self.total_number_of_files

        if total_size is None:
            if not size:
                return

            total_size, n = compute_directory_size(self.path)

        if total_size is not None:
            print(f"💽 Size       : {bytes(total_size)} ({number(total_size)})")
        if n is not None:
            print(f"📁 Files      : {number(n)}")

    @property
    def statistics(self):
        try:
            if self.dataset is not None:
                stats = self.dataset.statistics
                return stats["minimum"], stats["maximum"], stats["mean"], stats["stdev"]
        except AttributeError:
            return [["-"] * len(self.variables)] * 4

    @property
    def statistics_ready(self):
        for d in reversed(self.metadata.get("history", [])):
            if d["action"] == "compute_statistics_end":
                return True
        return False

    @property
    def statistics_started(self):
        for d in reversed(self.metadata.get("history", [])):
            if d["action"] == "compute_statistics_start":
                return datetime.datetime.fromisoformat(d["timestamp"])
        return None

    @property
    def build_flags(self):
        return self.zarr.get("_build_flags")

    @cached_property
    def copy_flags(self):
        if "_copy" not in self.zarr:
            return None
        return self.zarr["_copy"][:]

    @property
    def copy_in_progress(self):
        if "_copy" not in self.zarr:
            return False

        start = self.zarr["_copy"].attrs.get("copy_start_timestamp")
        end = self.zarr["_copy"].attrs.get("copy_end_timestamp")
        if start and end:
            return False

        return not all(self.copy_flags)

    @property
    def build_lengths(self):
        return self.zarr.get("_build_lengths")

    def progress(self):
        if self.copy_in_progress:
            copy_flags = self.copy_flags
            print("🪫  Dataset not ready, copy in progress.")
            assert isinstance(copy_flags, np.ndarray)
            total = len(copy_flags)
            built = copy_flags.sum()
            print(
                "📈 Progress:",
                progress(built, total, width=50),
                "{:.0f}%".format(built / total * 100),
            )
            return

        if self.build_flags is None:
            print("🪫 Dataset not initialized")
            return

        build_flags = self.build_flags

        build_lengths = self.build_lengths
        assert build_flags.size == build_lengths.size

        latest_write_timestamp = self.zarr.attrs.get("latest_write_timestamp")
        latest = datetime.datetime.fromisoformat(latest_write_timestamp) if latest_write_timestamp else None

        if not all(build_flags):
            if latest:
                print(f"🪫  Dataset not ready, last update {when(latest)}.")
            else:
                print("🪫  Dataset not ready.")
            total = sum(build_lengths)
            built = sum(ln if flag else 0 for ln, flag in zip(build_lengths, build_flags))
            print(
                "📈 Progress:",
                progress(built, total, width=50),
                "{:.0f}%".format(built / total * 100),
            )
            start = self.initialised
            if self.initialised:
                print(f"🕰️  Dataset initialized {when(start)}.")
                if built and latest:
                    speed = (latest - start) / built
                    eta = datetime.datetime.utcnow() + speed * (total - built)
                    print(f"🏁 ETA {when(eta)}.")
        else:
            if latest:
                print(f"🔋 Dataset ready, last update {when(latest)}.")
            else:
                print("🔋 Dataset ready.")
            if self.statistics_ready:
                print("📊 Statistics ready.")
            else:
                started = self.statistics_started
                if started:
                    print(f"⏳ Statistics not ready, started {when(started)}.")
                else:
                    print("⏳ Statistics not ready.")

    def brute_force_statistics(self):
        if self.dataset is None:
            return
        print("📊 Computing statistics...")
        # np.seterr(all="raise")

        nvars = self.dataset.shape[1]

        count = np.zeros(nvars, dtype=np.int64)
        sums = np.zeros(nvars, dtype=np.float32)
        squares = np.zeros(nvars, dtype=np.float32)

        minimum = np.full((nvars,), np.inf, dtype=np.float32)
        maximum = np.full((nvars,), -np.inf, dtype=np.float32)

        for i, chunk in enumerate(tqdm.tqdm(self.dataset, total=len(self.dataset))):
            values = chunk.reshape((nvars, -1))
            minimum = np.minimum(minimum, np.min(values, axis=1))
            maximum = np.maximum(maximum, np.max(values, axis=1))
            sums += np.sum(values, axis=1)
            squares += np.sum(values * values, axis=1)
            count += values.shape[1]

        mean = sums / count
        stats = [
            minimum,
            maximum,
            mean,
            np.sqrt(squares / count - mean * mean),
        ]

        rows = []

        for i, v in enumerate(self.variables):
            rows.append([i, v] + [x[i] for x in stats])

        print(
            table(
                rows,
                header=["Index", "Variable", "Min", "Max", "Mean", "Stdev"],
                align=[">", "<", ">", ">", ">", ">"],
                margin=3,
            )
        )


class NoVersion(Version):
    @property
    def first_date(self):
        monthly = find(self.metadata, "monthly")
        return datetime.datetime.fromisoformat(monthly["start"])

    @property
    def last_date(self):
        monthly = find(self.metadata, "monthly")
        time = max([int(t) for t in find(self.metadata["climetlab"], "time")])
        assert isinstance(time, int), (time, type(time))
        if time > 100:
            time = time // 100
        return datetime.datetime.fromisoformat(monthly["stop"]) + datetime.timedelta(hours=time)

    @property
    def frequency(self):
        time = find(self.metadata["climetlab"], "time")
        return 24 // len(time)

    @property
    def statistics(self):
        stats = find(self.metadata, "statistics_by_index")
        return stats["minimum"], stats["maximum"], stats["mean"], stats["stdev"]

    @property
    def statistics_ready(self):
        return find(self.metadata, "statistics_by_index") is not None

    @property
    def resolution(self):
        return find(self.metadata, "grid")

    def details(self):
        pass

    def progress(self):
        pass

    def ready(self):
        return True


class Version0_4(Version):
    def details(self):
        pass

    @property
    def initialised(self):
        return datetime.datetime.fromisoformat(self.metadata["creation_timestamp"])

    def statistics_ready(self):
        if not self.ready():
            return False
        build_flags = self.zarr["_build_flags"]
        return build_flags.attrs.get("_statistics_computed")

    def ready(self):
        if "_build_flags" not in self.zarr:
            return False

        build_flags = self.zarr["_build_flags"]
        if not build_flags.attrs.get("_initialised"):
            return False

        return all(build_flags)

    def _info(self, verbose, history, statistics, **kwargs):
        z = self.zarr

        # for backward compatibility
        if "climetlab" in z.attrs:
            climetlab_version = z.attrs["climetlab"].get("versions", {}).get("climetlab", "unkwown")
            print(f"climetlab version used to create this zarr: {climetlab_version}. Not supported.")
            return

        version = z.attrs.get("version")
        versions = z.attrs.get("versions")
        if not versions:
            print(" Cannot find metadata information about versions.")
        else:
            print(f"Zarr format (version {version})", end="")
            print(f" created by climetlab={versions.pop('climetlab')}", end="")
            timestamp = z.attrs.get("creation_timestamp")
            timestamp = datetime.datetime.fromisoformat(timestamp)
            print(f" on {timestamp}", end="")
            versions = ", ".join([f"{k}={v}" for k, v in versions.items()])
            print(f" using {versions}", end="")
            print()


class Version0_6(Version):
    @property
    def initialised(self):
        for record in self.metadata.get("history", []):
            if record["action"] == "initialised":
                return datetime.datetime.fromisoformat(record["timestamp"])

        # Sometimes the first record is missing
        timestamps = sorted([datetime.datetime.fromisoformat(d["timestamp"]) for d in self.metadata.get("history", [])])
        if timestamps:
            return timestamps[0]

        return None

    def details(self):
        print()
        for d in self.metadata.get("history", []):
            d = deepcopy(d)
            timestamp = d.pop("timestamp")
            timestamp = datetime.datetime.fromisoformat(timestamp)
            action = d.pop("action")
            versions = d.pop("versions")
            versions = ", ".join(f"{k}={v}" for k, v in versions.items())
            more = ", ".join(f"{k}={v}" for k, v in d.items())
            print(f"  {timestamp} : {action} ({versions}) {more}")
        print()

    def ready(self):
        if "_build_flags" not in self.zarr:
            return False

        build_flags = self.zarr["_build_flags"]
        return all(build_flags)

    @property
    def name_to_index(self):
        return {n: i for i, n in enumerate(self.metadata["variables"])}

    @property
    def variables(self):
        return self.metadata["variables"]


class Version0_12(Version0_6):
    def details(self):
        print()
        for d in self.metadata.get("history", []):
            d = deepcopy(d)
            timestamp = d.pop("timestamp")
            timestamp = datetime.datetime.fromisoformat(timestamp)
            action = d.pop("action")
            more = ", ".join(f"{k}={v}" for k, v in d.items())
            if more:
                more = f" ({more})"
            print(f"  {timestamp} : {action}{more}")
        print()

    @property
    def first_date(self):
        return datetime.datetime.fromisoformat(self.metadata["start_date"])

    @property
    def last_date(self):
        return datetime.datetime.fromisoformat(self.metadata["end_date"])


class Version0_13(Version0_12):
    @property
    def build_flags(self):
        if "_build" not in self.zarr:
            return None
        build = self.zarr["_build"]
        return build.get("flags")

    @property
    def build_lengths(self):
        if "_build" not in self.zarr:
            return None
        build = self.zarr["_build"]
        return build.get("lengths")


VERSIONS = {
    "0.0.0": NoVersion,
    "0.4.0": Version0_4,
    "0.6.0": Version0_6,
    "0.12.0": Version0_12,
    "0.13.0": Version0_13,
}


class InspectZarr:
    """Inspect a checkpoint or zarr file."""

    def inspect_zarr(self, path, **kwargs):
        version = self._info(path)

        # try:
        #     with open("/tmp/probe.json", "w") as f:
        #         json.dump(version.metadata, f, indent=4, sort_keys=True)
        # except Exception:
        #     pass

        dotted_line()
        version.describe()

        try:
            if kwargs.get("probe"):
                return version.probe()
            if kwargs.get("progress"):
                return version.progress()
            if kwargs.get("statistics"):
                return version.brute_force_statistics()
            version.info(kwargs.get("detailed"), kwargs.get("size"))
        except Exception as e:
            LOG.error("Error inspecting zarr file '%s': %s", path, e)

            print(type(version))
            raise

    def _info(self, path):
        if path.endswith("/"):
            path = path[:-1]

        try:
            z = open_zarr(path)
        except Exception as e:
            LOG.error("Error opening zarr file '%s': %s", path, e)
            raise

        metadata = dict(z.attrs)
        version = metadata.get("version", "0.0.0")
        if isinstance(version, int):
            version = f"0.{version}"

        version = semantic_version.Version.coerce(version)

        versions = {semantic_version.Version.coerce(k): v for k, v in VERSIONS.items()}

        candidate = None
        for v, klass in sorted(versions.items()):
            if version >= v:
                candidate = klass

        return candidate(path, z, metadata, version)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

from . import open_dataset
from .debug import Node
from .forewards import Forwards

LOG = logging.getLogger(__name__)


class Statistics(Forwards):
    def __init__(self, dataset, statistic):
        super().__init__(dataset)
        self._statistic = open_dataset(statistic)
        # TODO: relax that check to allow for a subset of variables
        if dataset.variables != self._statistic.variables:
            raise ValueError(
                f"Incompatible variables: {dataset.variables} and {self._statistic.variables} ({dataset} {self._statistic})"
            )

    @cached_property
    def statistics(self):
        return self._statistic.statistics

    def metadata_specific(self, **kwargs):
        return super().metadata_specific(
            statistics=self._statistic.metadata_specific(),
            **kwargs,
        )

    def tree(self):
        return Node(self, [self.forward.tree()])



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .forewards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Join(Combined):
    """Join the datasets along the variables axis."""

    def check_compatibility(self, d1, d2):
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=1)

    def check_same_variables(self, d1, d2):
        # Turned off because we are joining along the variables axis
        pass

    def __len__(self):
        return len(self.datasets[0])

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 1, slice(None))

        # TODO: optimize if index does not access all datasets, so we don't load chunks we don't need
        result = [d[index] for d in self.datasets]

        result = np.concatenate(result, axis=1)
        return apply_index_to_slices_changes(result[:, previous], changes)

    @debug_indexing
    def _get_slice(self, s):
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return np.concatenate([d[n] for d in self.datasets])

    @cached_property
    def shape(self):
        cols = sum(d.shape[1] for d in self.datasets)
        return (len(self), cols) + self.datasets[0].shape[2:]

    def _overlay(self):
        indices = {}
        i = 0
        for d in self.datasets:
            for v in d.variables:
                indices[v] = i
                i += 1

        if len(indices) == i:
            # No overlay
            return self

        variables = [v[1:-1] for v in self.variables if v[0] == "(" and v[-1] == ")"]
        indices = list(indices.values())

        i = 0
        for d in self.datasets:
            ok = False
            for v in d.variables:
                if i in indices:
                    ok = True
                i += 1
            if not ok:
                LOG.warning("Dataset %r completely overridden.", d)

        from .select import Select

        return Select(self, indices, {"overlay": variables})

    @cached_property
    def variables(self):
        seen = set()
        result = []
        for d in reversed(self.datasets):
            for v in reversed(d.variables):
                while v in seen:
                    v = f"({v})"
                seen.add(v)
                result.insert(0, v)

        return result

    @cached_property
    def name_to_index(self):
        return {k: i for i, k in enumerate(self.variables)}

    @property
    def statistics(self):
        return {
            k: np.concatenate([d.statistics[k] for d in self.datasets], axis=0) for k in self.datasets[0].statistics
        }

    def source(self, index):
        i = index
        for dataset in self.datasets:
            if i < dataset.shape[1]:
                return Source(self, index, dataset.source(i))
            i -= dataset.shape[1]
        assert False

    @cached_property
    def missing(self):
        result = set()
        for d in self.datasets:
            result = result | d.missing
        return result

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])


def join_factory(args, kwargs):

    datasets = kwargs.pop("join")
    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    assert isinstance(datasets, (list, tuple))

    datasets = [_open(e) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Join(datasets)._overlay()._subset(**kwargs)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import re
import textwrap
from functools import wraps

LOG = logging.getLogger(__name__)

TRACE_INDENT = 0


def step(action_path):
    return f"[{'.'.join(action_path)}]"


def trace(emoji, *args):
    if os.environ.get("ANEMOI_DATASET_TRACE_CREATE") is None:
        return
    print(emoji, " " * TRACE_INDENT, *args)


def trace_datasource(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        global TRACE_INDENT
        trace(
            "🌍",
            "=>",
            step(self.action_path),
            self._trace_datasource(*args, **kwargs),
        )
        TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        TRACE_INDENT -= 1
        trace(
            "🍎",
            "<=",
            step(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper


def trace_select(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        global TRACE_INDENT
        trace(
            "👓",
            "=>",
            ".".join(self.action_path),
            self._trace_select(*args, **kwargs),
        )
        TRACE_INDENT += 1
        result = method(self, *args, **kwargs)
        TRACE_INDENT -= 1
        trace(
            "🍍",
            "<=",
            ".".join(self.action_path),
            textwrap.shorten(repr(result), 256),
        )
        return result

    return wrapper


def notify_result(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(self, *args, **kwargs)
        self.context.notify_result(self.action_path, result)
        return result

    return wrapper


class Context:
    def __init__(self):
        # used_references is a set of reference paths that will be needed
        self.used_references = set()
        # results is a dictionary of reference path -> obj
        self.results = {}

    def will_need_reference(self, key):
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        self.used_references.add(key)

    def notify_result(self, key, result):
        trace("🎯", step(key), "notify result", result)
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.used_references:
            if key in self.results:
                raise ValueError(f"Duplicate result {key}")
            self.results[key] = result

    def get_result(self, key):
        assert isinstance(key, (list, tuple)), key
        key = tuple(key)
        if key in self.results:
            return self.results[key]
        all_keys = sorted(list(self.results.keys()))
        raise ValueError(f"Cannot find result {key} in {all_keys}")


class Substitution:
    pass


class Reference(Substitution):
    def __init__(self, context, action_path):
        self.context = context
        self.action_path = action_path

    def resolve(self, context):
        return context.get_result(self.action_path)


def resolve(context, x):
    if isinstance(x, tuple):
        return tuple([resolve(context, y) for y in x])

    if isinstance(x, list):
        return [resolve(context, y) for y in x]

    if isinstance(x, dict):
        return {k: resolve(context, v) for k, v in x.items()}

    if isinstance(x, Substitution):
        return x.resolve(context)

    return x


def substitute(context, x):
    if isinstance(x, tuple):
        return tuple([substitute(context, y) for y in x])

    if isinstance(x, list):
        return [substitute(context, y) for y in x]

    if isinstance(x, dict):
        return {k: substitute(context, v) for k, v in x.items()}

    if not isinstance(x, str):
        return x

    if re.match(r"^\${[\.\w]+}$", x):
        path = x[2:-1].split(".")
        context.will_need_reference(path)
        return Reference(context, path)

    return x



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
import logging
import shutil

import numpy as np

LOG = logging.getLogger(__name__)


def add_zarr_dataset(
    *,
    name,
    dtype=None,
    fill_value=None,
    zarr_root,
    shape=None,
    array=None,
    overwrite=True,
    **kwargs,
):
    if dtype is None:
        assert array is not None, (name, shape, array, dtype, zarr_root)
        dtype = array.dtype

    if shape is None:
        assert array is not None, (name, shape, array, dtype, zarr_root)
        shape = array.shape

    if array is not None:
        assert array.shape == shape, (array.shape, shape)
        a = zarr_root.create_dataset(
            name,
            shape=shape,
            dtype=dtype,
            overwrite=overwrite,
            **kwargs,
        )
        a[...] = array
        return a

    if "fill_value" not in kwargs:
        if str(dtype).startswith("float") or str(dtype).startswith("numpy.float"):
            kwargs["fill_value"] = np.nan
        elif str(dtype).startswith("datetime64") or str(dtype).startswith("numpy.datetime64"):
            kwargs["fill_value"] = np.datetime64("NaT")
        # elif str(dtype).startswith("timedelta64") or str(dtype).startswith(
        #    "numpy.timedelta64"
        # ):
        #    kwargs["fill_value"] = np.timedelta64("NaT")
        elif str(dtype).startswith("int") or str(dtype).startswith("numpy.int"):
            kwargs["fill_value"] = 0
        elif str(dtype).startswith("bool") or str(dtype).startswith("numpy.bool"):
            kwargs["fill_value"] = False
        else:
            raise ValueError(f"No fill_value for dtype={dtype}")

    a = zarr_root.create_dataset(
        name,
        shape=shape,
        dtype=dtype,
        overwrite=overwrite,
        **kwargs,
    )
    return a


class ZarrBuiltRegistry:
    name_lengths = "lengths"
    name_flags = "flags"
    lengths = None
    flags = None
    z = None

    def __init__(self, path, synchronizer_path=None):
        import zarr

        assert isinstance(path, str), path
        self.zarr_path = path

        if synchronizer_path is None:
            synchronizer_path = self.zarr_path + ".sync"
        self.synchronizer_path = synchronizer_path
        self.synchronizer = zarr.ProcessSynchronizer(self.synchronizer_path)

    def clean(self):
        try:
            shutil.rmtree(self.synchronizer_path)
        except FileNotFoundError:
            pass

    def _open_write(self):
        import zarr

        return zarr.open(self.zarr_path, mode="r+", synchronizer=self.synchronizer)

    def _open_read(self, sync=True):
        import zarr

        if sync:
            return zarr.open(self.zarr_path, mode="r", synchronizer=self.synchronizer)
        else:
            return zarr.open(self.zarr_path, mode="r")

    def new_dataset(self, *args, **kwargs):
        z = self._open_write()
        zarr_root = z["_build"]
        add_zarr_dataset(*args, zarr_root=zarr_root, overwrite=True, **kwargs)

    def add_to_history(self, action, **kwargs):
        new = dict(
            action=action,
            timestamp=datetime.datetime.utcnow().isoformat(),
        )
        new.update(kwargs)

        z = self._open_write()
        history = z.attrs.get("history", [])
        history.append(new)
        z.attrs["history"] = history

    def get_lengths(self):
        z = self._open_read()
        return list(z["_build"][self.name_lengths][:])

    def get_flags(self, **kwargs):
        z = self._open_read(**kwargs)
        return list(z["_build"][self.name_flags][:])

    def get_flag(self, i):
        z = self._open_read()
        return z["_build"][self.name_flags][i]

    def set_flag(self, i, value=True):
        z = self._open_write()
        z.attrs["latest_write_timestamp"] = datetime.datetime.utcnow().isoformat()
        z["_build"][self.name_flags][i] = value

    def create(self, lengths, overwrite=False):
        self.new_dataset(name=self.name_lengths, array=np.array(lengths, dtype="i4"))
        self.new_dataset(name=self.name_flags, array=np.array([False] * len(lengths), dtype=bool))
        self.add_to_history("initialised")

    def reset(self, lengths):
        return self.create(lengths, overwrite=True)

    def add_provenance(self, name):
        from anemoi.utils.provenance import gather_provenance_info

        z = self._open_write()
        z.attrs[name] = gather_provenance_info()



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import warnings
from copy import deepcopy

import numpy as np
from climetlab.core.temporary import temp_file
from climetlab.readers.grib.output import new_grib_output

from anemoi.datasets.create.check import check_data_values
from anemoi.datasets.create.functions import assert_is_fieldset
from anemoi.datasets.create.functions.actions.mars import mars


def to_list(x):
    if isinstance(x, (list, tuple)):
        return x
    if isinstance(x, str):
        return x.split("/")
    return [x]


def normalise_number(number):
    number = to_list(number)

    if len(number) > 4 and (number[1] == "to" and number[3] == "by"):
        return list(range(int(number[0]), int(number[2]) + 1, int(number[4])))

    if len(number) > 2 and number[1] == "to":
        return list(range(int(number[0]), int(number[2]) + 1))

    return number


def normalise_request(request):
    request = deepcopy(request)
    if "number" in request:
        request["number"] = normalise_number(request["number"])
    if "time" in request:
        request["time"] = to_list(request["time"])
    request["param"] = to_list(request["param"])
    return request


def load_if_needed(context, dates, dict_or_dataset):
    if isinstance(dict_or_dataset, dict):
        dict_or_dataset = normalise_request(dict_or_dataset)
        dict_or_dataset = mars(context, dates, dict_or_dataset)
    return dict_or_dataset


def perturbations(context, dates, members, center, remapping={}, patches={}):
    members = load_if_needed(context, dates, members)
    center = load_if_needed(context, dates, center)

    keys = ["param", "level", "valid_datetime", "date", "time", "step", "number"]

    def check_compatible(f1, f2, ignore=["number"]):
        for k in keys + ["grid", "shape"]:
            if k in ignore:
                continue
            assert f1.metadata(k) == f2.metadata(k), (k, f1.metadata(k), f2.metadata(k))

    print(f"Retrieving ensemble data with {members}")
    print(f"Retrieving center data with {center}")

    members = members.order_by(*keys)
    center = center.order_by(*keys)

    number_list = members.unique_values("number")["number"]
    n_numbers = len(number_list)

    if len(center) * n_numbers != len(members):
        print(len(center), n_numbers, len(members))
        for f in members:
            print("Member: ", f)
        for f in center:
            print("Center: ", f)
        raise ValueError(f"Inconsistent number of fields: {len(center)} * {n_numbers} != {len(members)}")

    # prepare output tmp file so we can read it back
    tmp = temp_file()
    path = tmp.path
    out = new_grib_output(path)

    for i, center_field in enumerate(center):
        param = center_field.metadata("param")

        # load the center field
        center_np = center_field.to_numpy()

        # load the ensemble fields and compute the mean
        members_np = np.zeros((n_numbers, *center_np.shape))

        for j in range(n_numbers):
            ensemble_field = members[i * n_numbers + j]
            check_compatible(center_field, ensemble_field)
            members_np[j] = ensemble_field.to_numpy()

        mean_np = members_np.mean(axis=0)

        for j in range(n_numbers):
            template = members[i * n_numbers + j]
            e = members_np[j]
            m = mean_np
            c = center_np

            assert e.shape == c.shape == m.shape, (e.shape, c.shape, m.shape)

            FORCED_POSITIVE = [
                "q",
                "cp",
                "lsp",
                "tp",
            ]  # add "swl4", "swl3", "swl2", "swl1", "swl0", and more ?

            x = c - m + e

            if param in FORCED_POSITIVE:
                warnings.warn(f"Clipping {param} to be positive")
                x = np.maximum(x, 0)

            assert x.shape == e.shape, (x.shape, e.shape)

            check_data_values(x, name=param)
            out.write(x, template=template)
            template = None

    out.close()

    from climetlab import load_source

    ds = load_source("file", path)
    assert_is_fieldset(ds)
    # save a reference to the tmp file so it is deleted
    # only when the dataset is not used anymore
    ds._tmp = tmp

    assert len(ds) == len(members), (len(ds), len(members))

    return ds


execute = perturbations



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import warnings

import numpy as np

LOG = logging.getLogger(__name__)


class CubesFilter:
    def __init__(self, *, parts, total):
        if parts is None:
            self.parts = None
            return

        if len(parts) == 1:
            part = parts[0]
            if part.lower() in ["all", "*"]:
                self.parts = None
                return

            if "/" in part:
                i_chunk, n_chunks = part.split("/")
                i_chunk, n_chunks = int(i_chunk), int(n_chunks)

                assert i_chunk > 0, f"Chunk number {i_chunk} must be positive."
                if n_chunks > total:
                    warnings.warn(
                        f"Number of chunks {n_chunks} is larger than the total number of chunks: {total}+1. "
                        "Some chunks will be empty."
                    )

                chunk_size = total / n_chunks
                parts = [x for x in range(total) if x >= (i_chunk - 1) * chunk_size and x < i_chunk * chunk_size]

        parts = [int(_) for _ in parts]
        LOG.info(f"Running parts: {parts}")
        if not parts:
            warnings.warn(f"Nothing to do for chunk {i_chunk}/{n_chunks}.")

        self.parts = parts

    def __call__(self, i):
        if self.parts is None:
            return True
        return i in self.parts


class ViewCacheArray:
    """A class that provides a caching mechanism for writing to a NumPy-like array.

    The is initialized with a NumPy-like array, a shape and a list to reindex the first dimension.
    The array is used to store the final data, while the cache is used to temporarily
    store the data before flushing it to the array.

    The `flush` method copies the contents of the cache to the final array.
    """

    def __init__(self, array, *, shape, indexes):
        assert len(indexes) == shape[0], (len(indexes), shape[0])
        self.array = array
        self.dtype = array.dtype
        self.cache = np.full(shape, np.nan, dtype=self.dtype)
        self.indexes = indexes

    def __setitem__(self, key, value):
        self.cache[key] = value

    def flush(self):
        for i in range(self.cache.shape[0]):
            global_i = self.indexes[i]
            self.array[global_i] = self.cache[i]



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import warnings
from functools import cached_property

import numpy as np
import zarr

from . import MissingDateError
from .dataset import Dataset
from .debug import DEBUG_ZARR_LOADING
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .indexing import expand_list_indexing
from .misc import load_config

LOG = logging.getLogger(__name__)


class ReadOnlyStore(zarr.storage.BaseStore):
    def __delitem__(self, key):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()


class HTTPStore(ReadOnlyStore):
    """We write our own HTTPStore because the one used by zarr (fsspec) does not play
    well with fork() and multiprocessing."""

    def __init__(self, url):
        self.url = url

    def __getitem__(self, key):
        import requests

        r = requests.get(self.url + "/" + key)

        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return r.content


class S3Store(ReadOnlyStore):
    """We write our own S3Store because the one used by zarr (fsspec) does not play well
    with fork() and multiprocessing."""

    def __init__(self, url):
        import boto3

        self.bucket, self.key = url[5:].split("/", 1)

        # TODO: get the profile name from the url
        self.s3 = boto3.Session(profile_name=None).client("s3")

    def __getitem__(self, key):
        try:
            response = self.s3.get_object(Bucket=self.bucket, Key=self.key + "/" + key)
        except self.s3.exceptions.NoSuchKey:
            raise KeyError(key)

        return response["Body"].read()


class DebugStore(ReadOnlyStore):
    def __init__(self, store):
        assert not isinstance(store, DebugStore)
        self.store = store

    def __getitem__(self, key):
        # print()
        print("GET", key, self)
        # traceback.print_stack(file=sys.stdout)
        return self.store[key]

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        warnings.warn("DebugStore: iterating over the store")
        return iter(self.store)

    def __contains__(self, key):
        return key in self.store


def open_zarr(path, dont_fail=False):
    try:
        store = path

        if store.startswith("http://") or store.startswith("https://"):
            store = HTTPStore(store)

        elif store.startswith("s3://"):
            store = S3Store(store)

        if DEBUG_ZARR_LOADING:
            if isinstance(store, str):
                store = zarr.storage.DirectoryStore(store)
            store = DebugStore(store)

        return zarr.convenience.open(store, "r")
    except zarr.errors.PathNotFoundError:
        if not dont_fail:
            raise zarr.errors.PathNotFoundError(path)


class Zarr(Dataset):
    def __init__(self, path):
        if isinstance(path, zarr.hierarchy.Group):
            self.was_zarr = True
            self.path = str(id(path))
            self.z = path
        else:
            self.was_zarr = False
            self.path = str(path)
            self.z = open_zarr(self.path)

        # This seems to speed up the reading of the data a lot
        self.data = self.z.data
        self.missing = set()

    @classmethod
    def from_name(cls, name):
        if name.endswith(".zip") or name.endswith(".zarr"):
            return Zarr(name)
        return Zarr(zarr_lookup(name))

    def __len__(self):
        return self.data.shape[0]

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n):
        return self.data[n]

    def _unwind(self, index, rest, shape, axis, axes):
        if not isinstance(index, (int, slice, list, tuple)):
            try:
                # NumPy arrays, TensorFlow tensors, etc.
                index = tuple(index.tolist())
                assert not isinstance(index, bool), "Mask not supported"
            except AttributeError:
                pass

        if isinstance(index, (list, tuple)):
            axes.append(axis)  # Dimension of the concatenation
            for i in index:
                yield from self._unwind((slice(i, i + 1),), rest, shape, axis, axes)
            return

        if len(rest) == 0:
            yield (index,)
            return

        for n in self._unwind(rest[0], rest[1:], shape, axis + 1, axes):
            yield (index,) + n

    @cached_property
    def chunks(self):
        return self.z.data.chunks

    @cached_property
    def shape(self):
        return self.data.shape

    @cached_property
    def dtype(self):
        return self.z.data.dtype

    @cached_property
    def dates(self):
        return self.z.dates[:]  # Convert to numpy

    @property
    def latitudes(self):
        try:
            return self.z.latitudes[:]
        except AttributeError:
            LOG.warning("No 'latitudes' in %r, trying 'latitude'", self)
            return self.z.latitude[:]

    @property
    def longitudes(self):
        try:
            return self.z.longitudes[:]
        except AttributeError:
            LOG.warning("No 'longitudes' in %r, trying 'longitude'", self)
            return self.z.longitude[:]

    @property
    def statistics(self):
        return dict(
            mean=self.z.mean[:],
            stdev=self.z.stdev[:],
            maximum=self.z.maximum[:],
            minimum=self.z.minimum[:],
        )

    @property
    def resolution(self):
        return self.z.attrs["resolution"]

    @property
    def field_shape(self):
        return tuple(self.z.attrs["field_shape"])

    @property
    def frequency(self):
        try:
            return self.z.attrs["frequency"]
        except KeyError:
            LOG.warning("No 'frequency' in %r, computing from 'dates'", self)
        dates = self.dates
        delta = dates[1].astype(object) - dates[0].astype(object)
        return int(delta.total_seconds() / 3600)

    @property
    def name_to_index(self):
        if "variables" in self.z.attrs:
            return {n: i for i, n in enumerate(self.z.attrs["variables"])}
        return self.z.attrs["name_to_index"]

    @property
    def variables(self):
        return [
            k
            for k, v in sorted(
                self.name_to_index.items(),
                key=lambda x: x[1],
            )
        ]

    def __repr__(self):
        return self.path

    def end_of_statistics_date(self):
        return self.dates[-1]

    def metadata_specific(self):
        return super().metadata_specific(
            attrs=dict(self.z.attrs),
            chunks=self.chunks,
            dtype=str(self.dtype),
        )

    def source(self, index):
        return Source(self, index, info=self.path)

    def mutate(self):
        if len(self.z.attrs.get("missing_dates", [])):
            LOG.warning(f"Dataset {self} has missing dates")
            return ZarrWithMissingDates(self.z if self.was_zarr else self.path)
        return self

    def tree(self):
        return Node(self, [], path=self.path)


class ZarrWithMissingDates(Zarr):
    def __init__(self, path):
        super().__init__(path)

        missing_dates = self.z.attrs.get("missing_dates", [])
        missing_dates = [np.datetime64(x) for x in missing_dates]
        self.missing_to_dates = {i: d for i, d in enumerate(self.dates) if d in missing_dates}
        self.missing = set(self.missing_to_dates)

    def mutate(self):
        return self

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n):
        if isinstance(n, int):
            if n in self.missing:
                self._report_missing(n)
            return self.data[n]

        if isinstance(n, slice):
            common = set(range(*n.indices(len(self)))) & self.missing
            if common:
                self._report_missing(list(common)[0])
            return self.data[n]

        if isinstance(n, tuple):
            first = n[0]
            if isinstance(first, int):
                if first in self.missing:
                    self._report_missing(first)
                return self.data[n]

            if isinstance(first, slice):
                common = set(range(*first.indices(len(self)))) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

            if isinstance(first, (list, tuple)):
                common = set(first) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.data[n]

        raise TypeError(f"Unsupported index {n} {type(n)}")

    def _report_missing(self, n):
        raise MissingDateError(f"Date {self.missing_to_dates[n]} is missing (index={n})")

    def tree(self):
        return Node(self, [], path=self.path, missing=sorted(self.missing))


def zarr_lookup(name):

    if name.endswith(".zarr") or name.endswith(".zip"):
        return name

    config = load_config()["datasets"]

    if name in config["named"]:
        return config["named"][name]

    tried = []
    for location in config["path"]:
        if not location.endswith("/"):
            location += "/"
        full = location + name + ".zarr"
        tried.append(full)
        try:
            z = open_zarr(full, dont_fail=True)
            if z is not None:
                # Cache for next time
                config["named"][name] = full
                return full
        except zarr.errors.PathNotFoundError:
            pass

    raise ValueError(f"Cannot find a dataset that matched '{name}'. Tried: {tried}")



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
import datetime
import logging
import os
import time
import uuid
from functools import cached_property

import numpy as np
import zarr

from anemoi.datasets import open_dataset
from anemoi.datasets.data.misc import as_first_date
from anemoi.datasets.data.misc import as_last_date
from anemoi.datasets.dates.groups import Groups

from .check import DatasetName
from .check import check_data_values
from .config import build_output
from .config import loader_config
from .input import build_input
from .statistics import TempStatistics
from .statistics import compute_statistics
from .utils import bytes
from .utils import compute_directory_sizes
from .utils import normalize_and_check_dates
from .utils import progress_bar
from .utils import seconds
from .writer import CubesFilter
from .writer import ViewCacheArray
from .zarr import ZarrBuiltRegistry
from .zarr import add_zarr_dataset

LOG = logging.getLogger(__name__)

VERSION = "0.20"


def default_statistics_dates(dates):
    """
    Calculate default statistics dates based on the given list of dates.

    Args:
        dates (list): List of datetime objects representing dates.

    Returns:
        tuple: A tuple containing the default start and end dates.
    """
    first = dates[0]
    last = dates[-1]
    if isinstance(first, np.datetime64):
        first = first.tolist()
    if isinstance(last, np.datetime64):
        last = last.tolist()
    assert isinstance(first, datetime.datetime), first
    assert isinstance(last, datetime.datetime), last

    n_years = round((last - first).total_seconds() / (365.25 * 24 * 60 * 60))

    if n_years < 10:
        # leave out 20% of the data
        k = int(len(dates) * 0.8)
        end = dates[k - 1]
        LOG.info(f"Number of years {n_years} < 10, leaving out 20%. {end=}")
        return dates[0], end

    delta = 1
    if n_years >= 20:
        delta = 3
    LOG.info(f"Number of years {n_years}, leaving out {delta} years.")
    end_year = last.year - delta
    end = max(d for d in dates if d.year == end_year)
    return dates[0], end


class Loader:
    def __init__(self, *, path, print=print, **kwargs):
        # Catch all floating point errors, including overflow, sqrt(<0), etc
        np.seterr(all="raise", under="warn")

        assert isinstance(path, str), path

        self.path = path
        self.kwargs = kwargs
        self.print = print

        statistics_tmp = kwargs.get("statistics_tmp") or self.path + ".statistics"

        self.statistics_registry = TempStatistics(statistics_tmp)

    @classmethod
    def from_config(cls, *, config, path, print=print, **kwargs):
        # config is the path to the config file or a dict with the config
        assert isinstance(config, dict) or isinstance(config, str), config
        return cls(config=config, path=path, print=print, **kwargs)

    @classmethod
    def from_dataset_config(cls, *, path, print=print, **kwargs):
        assert os.path.exists(path), f"Path {path} does not exist."
        z = zarr.open(path, mode="r")
        config = z.attrs["_create_yaml_config"]
        LOG.info(f"Config loaded from zarr config: {config}")
        return cls.from_config(config=config, path=path, print=print, **kwargs)

    @classmethod
    def from_dataset(cls, *, path, **kwargs):
        assert os.path.exists(path), f"Path {path} does not exist."
        return cls(path=path, **kwargs)

    def build_input(self):
        from climetlab.core.order import build_remapping

        builder = build_input(
            self.main_config.input,
            data_sources=self.main_config.get("data_sources", {}),
            order_by=self.output.order_by,
            flatten_grid=self.output.flatten_grid,
            remapping=build_remapping(self.output.remapping),
        )
        LOG.info("✅ INPUT_BUILDER")
        LOG.info(builder)
        return builder

    def build_statistics_dates(self, start, end):
        ds = open_dataset(self.path)
        dates = ds.dates

        default_start, default_end = default_statistics_dates(dates)
        if start is None:
            start = default_start
        if end is None:
            end = default_end

        start = as_first_date(start, dates)
        end = as_last_date(end, dates)

        start = start.astype(datetime.datetime)
        end = end.astype(datetime.datetime)
        return (start.isoformat(), end.isoformat())

    def read_dataset_metadata(self):
        ds = open_dataset(self.path)
        self.dataset_shape = ds.shape
        self.variables_names = ds.variables
        assert len(self.variables_names) == ds.shape[1], self.dataset_shape
        self.dates = ds.dates

        z = zarr.open(self.path, "r")
        self.missing_dates = z.attrs.get("missing_dates", [])
        self.missing_dates = [np.datetime64(d) for d in self.missing_dates]

    def allow_nan(self, name):
        return name in self.main_config.statistics.get("allow_nans", [])

    @cached_property
    def registry(self):
        return ZarrBuiltRegistry(self.path)

    def initialise_dataset_backend(self):
        z = zarr.open(self.path, mode="w")
        z.create_group("_build")

    def update_metadata(self, **kwargs):
        LOG.info(f"Updating metadata {kwargs}")
        z = zarr.open(self.path, mode="w+")
        for k, v in kwargs.items():
            if isinstance(v, np.datetime64):
                v = v.astype(datetime.datetime)
            if isinstance(v, datetime.date):
                v = v.isoformat()
            z.attrs[k] = v

    def _add_dataset(self, mode="r+", **kwargs):
        z = zarr.open(self.path, mode=mode)
        return add_zarr_dataset(zarr_root=z, **kwargs)

    def get_zarr_chunks(self):
        z = zarr.open(self.path, mode="r")
        return z["data"].chunks

    def print_info(self):
        z = zarr.open(self.path, mode="r")
        try:
            LOG.info(z["data"].info)
        except Exception as e:
            LOG.info(e)


class InitialiseLoader(Loader):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.main_config = loader_config(config)

        self.statistics_registry.delete()

        LOG.info(self.main_config.dates)
        self.groups = Groups(**self.main_config.dates)

        self.output = build_output(self.main_config.output, parent=self)
        self.input = self.build_input()

        LOG.info(self.input)
        all_dates = self.groups.dates
        self.minimal_input = self.input.select([all_dates[0]])

        LOG.info(self.groups)
        LOG.info("MINIMAL INPUT :")
        LOG.info(self.minimal_input)

    def initialise(self, check_name=True):
        """Create empty dataset."""

        self.print("Config loaded ok:")
        LOG.info(self.main_config)

        dates = self.groups.dates
        frequency = dates.frequency
        assert isinstance(frequency, int), frequency

        self.print(f"Found {len(dates)} datetimes.")
        LOG.info(f"Dates: Found {len(dates)} datetimes, in {len(self.groups)} groups: ")
        LOG.info(f"Missing dates: {len(dates.missing)}")
        lengths = [len(g) for g in self.groups]
        self.print(f"Found {len(dates)} datetimes {'+'.join([str(_) for _ in lengths])}.")

        variables = self.minimal_input.variables
        self.print(f"Found {len(variables)} variables : {','.join(variables)}.")

        variables_with_nans = self.main_config.statistics.get("allow_nans", [])

        ensembles = self.minimal_input.ensembles
        self.print(f"Found {len(ensembles)} ensembles : {','.join([str(_) for _ in ensembles])}.")

        grid_points = self.minimal_input.grid_points
        LOG.info(f"gridpoints size: {[len(i) for i in grid_points]}")

        resolution = self.minimal_input.resolution
        LOG.info(f"{resolution=}")

        coords = self.minimal_input.coords
        coords["dates"] = dates
        total_shape = self.minimal_input.shape
        total_shape[0] = len(dates)
        self.print(f"total_shape = {total_shape}")

        chunks = self.output.get_chunking(coords)
        LOG.info(f"{chunks=}")
        dtype = self.output.dtype

        self.print(f"Creating Dataset '{self.path}', with {total_shape=}, {chunks=} and {dtype=}")

        metadata = {}
        metadata["uuid"] = str(uuid.uuid4())

        metadata.update(self.main_config.get("add_metadata", {}))

        metadata["_create_yaml_config"] = self.main_config.get_serialisable_dict()

        metadata["description"] = self.main_config.description
        metadata["version"] = VERSION

        metadata["data_request"] = self.minimal_input.data_request
        metadata["remapping"] = self.output.remapping

        metadata["order_by"] = self.output.order_by_as_list
        metadata["flatten_grid"] = self.output.flatten_grid

        metadata["ensemble_dimension"] = len(ensembles)
        metadata["variables"] = variables
        metadata["variables_with_nans"] = variables_with_nans
        metadata["resolution"] = resolution
        metadata["field_shape"] = self.minimal_input.field_shape
        metadata["proj_string"] = self.minimal_input.proj_string

        metadata["licence"] = self.main_config["licence"]
        metadata["attribution"] = self.main_config["attribution"]

        metadata["frequency"] = frequency
        metadata["start_date"] = dates[0].isoformat()
        metadata["end_date"] = dates[-1].isoformat()
        metadata["missing_dates"] = [_.isoformat() for _ in dates.missing]

        if check_name:
            basename, ext = os.path.splitext(os.path.basename(self.path))  # noqa: F841
            ds_name = DatasetName(
                basename,
                resolution,
                dates[0],
                dates[-1],
                frequency,
            )
            ds_name.raise_if_not_valid(print=self.print)

        if len(dates) != total_shape[0]:
            raise ValueError(
                f"Final date size {len(dates)} (from {dates[0]} to {dates[-1]}, {frequency=}) "
                f"does not match data shape {total_shape[0]}. {total_shape=}"
            )

        dates = normalize_and_check_dates(
            dates,
            metadata["start_date"],
            metadata["end_date"],
            metadata["frequency"],
        )

        metadata.update(self.main_config.get("force_metadata", {}))

        ###############################################################
        # write data
        ###############################################################

        self.initialise_dataset_backend()

        self.update_metadata(**metadata)

        self._add_dataset(name="data", chunks=chunks, dtype=dtype, shape=total_shape)
        self._add_dataset(name="dates", array=dates)
        self._add_dataset(name="latitudes", array=grid_points[0])
        self._add_dataset(name="longitudes", array=grid_points[1])

        self.registry.create(lengths=lengths)
        self.statistics_registry.create(exist_ok=False)
        self.registry.add_to_history("statistics_registry_initialised", version=self.statistics_registry.version)

        statistics_start, statistics_end = self.build_statistics_dates(
            self.main_config.statistics.get("start"),
            self.main_config.statistics.get("end"),
        )
        self.update_metadata(
            statistics_start_date=statistics_start,
            statistics_end_date=statistics_end,
        )
        LOG.info(f"Will compute statistics from {statistics_start} to {statistics_end}")

        self.registry.add_to_history("init finished")

        assert chunks == self.get_zarr_chunks(), (chunks, self.get_zarr_chunks())


class ContentLoader(Loader):
    def __init__(self, config, parts, **kwargs):
        super().__init__(**kwargs)
        self.main_config = loader_config(config)

        self.groups = Groups(**self.main_config.dates)
        self.output = build_output(self.main_config.output, parent=self)
        self.input = self.build_input()
        self.read_dataset_metadata()

        self.parts = parts
        total = len(self.registry.get_flags())
        self.cube_filter = CubesFilter(parts=self.parts, total=total)

        self.data_array = zarr.open(self.path, mode="r+")["data"]
        self.n_groups = len(self.groups)

    def load(self):
        self.registry.add_to_history("loading_data_start", parts=self.parts)

        for igroup, group in enumerate(self.groups):
            if not self.cube_filter(igroup):
                continue
            if self.registry.get_flag(igroup):
                LOG.info(f" -> Skipping {igroup} total={len(self.groups)} (already done)")
                continue
            self.print(f" -> Processing {igroup} total={len(self.groups)}")
            # print("========", group)
            assert isinstance(group[0], datetime.datetime), group

            result = self.input.select(dates=group)
            assert result.dates == group, (len(result.dates), len(group))

            msg = f"Building data for group {igroup}/{self.n_groups}"
            LOG.info(msg)
            self.print(msg)

            # There are several groups.
            # There is one result to load for each group.
            self.load_result(result)
            self.registry.set_flag(igroup)

        self.registry.add_to_history("loading_data_end", parts=self.parts)
        self.registry.add_provenance(name="provenance_load")
        self.statistics_registry.add_provenance(name="provenance_load", config=self.main_config)

        self.print_info()

    def load_result(self, result):
        # There is one cube to load for each result.
        dates = result.dates

        cube = result.get_cube()
        assert cube.extended_user_shape[0] == len(dates), (
            cube.extended_user_shape[0],
            len(dates),
        )

        shape = cube.extended_user_shape
        dates_in_data = cube.user_coords["valid_datetime"]

        LOG.info(f"Loading {shape=} in {self.data_array.shape=}")

        def check_dates_in_data(lst, lst2):
            lst2 = [np.datetime64(_) for _ in lst2]
            lst = [np.datetime64(_) for _ in lst]
            assert lst == lst2, ("Dates in data are not the requested ones:", lst, lst2)

        check_dates_in_data(dates_in_data, dates)

        def dates_to_indexes(dates, all_dates):
            x = np.array(dates, dtype=np.datetime64)
            y = np.array(all_dates, dtype=np.datetime64)
            bitmap = np.isin(x, y)
            return np.where(bitmap)[0]

        indexes = dates_to_indexes(self.dates, dates_in_data)

        array = ViewCacheArray(self.data_array, shape=shape, indexes=indexes)
        self.load_cube(cube, array)

        stats = compute_statistics(array.cache, self.variables_names, allow_nan=self.allow_nan)
        self.statistics_registry.write(indexes, stats, dates=dates_in_data)

        array.flush()

    def load_cube(self, cube, array):
        # There are several cubelets for each cube
        start = time.time()
        load = 0
        save = 0

        reading_chunks = None
        total = cube.count(reading_chunks)
        self.print(f"Loading datacube: {cube}")
        bar = progress_bar(
            iterable=cube.iterate_cubelets(reading_chunks),
            total=total,
            desc=f"Loading datacube {cube}",
        )
        for i, cubelet in enumerate(bar):
            bar.set_description(f"Loading {i}/{total}")

            now = time.time()
            data = cubelet.to_numpy()
            local_indexes = cubelet.coords
            load += time.time() - now

            name = self.variables_names[local_indexes[1]]
            check_data_values(
                data[:],
                name=name,
                log=[i, data.shape, local_indexes],
                allow_nan=self.allow_nan,
            )

            now = time.time()
            array[local_indexes] = data
            save += time.time() - now

        now = time.time()
        save += time.time() - now
        LOG.info("Written.")
        msg = f"Elapsed: {seconds(time.time() - start)}, load time: {seconds(load)}, write time: {seconds(save)}."
        self.print(msg)
        LOG.info(msg)


class StatisticsLoader(Loader):
    main_config = {}

    def __init__(
        self,
        config=None,
        statistics_output=None,
        statistics_start=None,
        statistics_end=None,
        force=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.user_statistics_start = statistics_start
        self.user_statistics_end = statistics_end

        self.statistics_output = statistics_output

        self.output_writer = {
            None: self.write_stats_to_dataset,
            "-": self.write_stats_to_stdout,
        }.get(self.statistics_output, self.write_stats_to_file)

        if config:
            self.main_config = loader_config(config)

        self.read_dataset_metadata()

    def _get_statistics_dates(self):
        dates = self.dates
        dtype = type(dates[0])

        def assert_dtype(d):
            assert type(d) is dtype, (type(d), dtype)

        # remove missing dates
        if self.missing_dates:
            assert type(self.missing_dates[0]) is dtype, (
                type(self.missing_dates[0]),
                dtype,
            )
        dates = [d for d in dates if d not in self.missing_dates]

        # filter dates according the the start and end dates in the metadata
        z = zarr.open(self.path, mode="r")
        start, end = z.attrs.get("statistics_start_date"), z.attrs.get("statistics_end_date")
        start, end = np.datetime64(start), np.datetime64(end)
        assert_dtype(start)
        dates = [d for d in dates if d >= start and d <= end]

        # filter dates according the the user specified start and end dates
        if self.user_statistics_start:
            limit = as_first_date(self.user_statistics_start, dates)
            limit = np.datetime64(limit)
            assert_dtype(limit)
            dates = [d for d in dates if d >= limit]

        if self.user_statistics_end:
            limit = as_last_date(self.user_statistics_start, dates)
            limit = np.datetime64(limit)
            assert_dtype(limit)
            dates = [d for d in dates if d <= limit]

        return dates

    def run(self):
        dates = self._get_statistics_dates()
        stats = self.statistics_registry.get_aggregated(dates, self.variables_names, self.allow_nan)
        self.output_writer(stats)

    def write_stats_to_file(self, stats):
        stats.save(self.statistics_output, provenance=dict(config=self.main_config))
        LOG.info(f"✅ Statistics written in {self.statistics_output}")

    def write_stats_to_dataset(self, stats):
        if self.user_statistics_start or self.user_statistics_end:
            raise ValueError(
                (
                    "Cannot write statistics in dataset with user specified dates. "
                    "This would be conflicting with the dataset metadata."
                )
            )

        if not all(self.registry.get_flags(sync=False)):
            raise Exception(f"❗Zarr {self.path} is not fully built, not writting statistics into dataset.")

        for k in ["mean", "stdev", "minimum", "maximum", "sums", "squares", "count", "has_nans"]:
            self._add_dataset(name=k, array=stats[k])

        self.registry.add_to_history("compute_statistics_end")
        LOG.info(f"Wrote statistics in {self.path}")

    def write_stats_to_stdout(self, stats):
        LOG.info(stats)


class SizeLoader(Loader):
    def __init__(self, path, print):
        self.path = path
        self.print = print

    def add_total_size(self):
        dic = compute_directory_sizes(self.path)

        size = dic["total_size"]
        n = dic["total_number_of_files"]

        LOG.info(f"Total size: {bytes(size)}")
        LOG.info(f"Total number of files: {n}")

        self.update_metadata(total_size=size, total_number_of_files=n)


class CleanupLoader(Loader):
    def run(self):
        self.statistics_registry.delete()
        self.registry.clean()



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import os

from .. import Command

# from .checkpoint import InspectCheckpoint
from .zarr import InspectZarr


class Inspect(Command, InspectZarr):
    # class Inspect(Command, InspectCheckpoint, InspectZarr):
    """Inspect a checkpoint or zarr file."""

    def add_arguments(self, command_parser):
        # g = command_parser.add_mutually_exclusive_group()
        # g.add_argument("--inspect", action="store_true", help="Inspect weights")
        command_parser.add_argument("path", metavar="PATH", nargs="+")
        command_parser.add_argument("--detailed", action="store_true")
        # command_parser.add_argument("--probe", action="store_true")
        command_parser.add_argument("--progress", action="store_true")
        command_parser.add_argument("--statistics", action="store_true")
        command_parser.add_argument("--size", action="store_true", help="Print size")

    def run(self, args):
        dic = vars(args)
        for path in dic.pop("path"):
            if os.path.isdir(path) or path.endswith(".zarr.zip") or path.endswith(".zarr"):
                self.inspect_zarr(path=path, **dic)
            else:
                raise ValueError(f"Unknown file type: {path}")
                # self.inspect_checkpoint(path=path, **dic)


command = Inspect



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from .dataset import Dataset
from .debug import debug_indexing
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class Forwards(Dataset):
    def __init__(self, forward):
        self.forward = forward

    def __len__(self):
        return len(self.forward)

    def __getitem__(self, n):
        return self.forward[n]

    @property
    def dates(self):
        return self.forward.dates

    @property
    def resolution(self):
        return self.forward.resolution

    @property
    def field_shape(self):
        return self.forward.field_shape

    @property
    def frequency(self):
        return self.forward.frequency

    @property
    def latitudes(self):
        return self.forward.latitudes

    @property
    def longitudes(self):
        return self.forward.longitudes

    @property
    def name_to_index(self):
        return self.forward.name_to_index

    @property
    def variables(self):
        return self.forward.variables

    @property
    def statistics(self):
        return self.forward.statistics

    @property
    def shape(self):
        return self.forward.shape

    @property
    def dtype(self):
        return self.forward.dtype

    @property
    def missing(self):
        return self.forward.missing

    @property
    def grids(self):
        return self.forward.grids

    def metadata_specific(self, **kwargs):
        return super().metadata_specific(
            forward=self.forward.metadata_specific(),
            **kwargs,
        )

    def source(self, index):
        return self.forward.source(index)


class Combined(Forwards):
    def __init__(self, datasets):
        self.datasets = datasets
        assert len(self.datasets) > 1, len(self.datasets)

        for d in self.datasets[1:]:
            self.check_compatibility(self.datasets[0], d)

        # Forward most properties to the first dataset
        super().__init__(datasets[0])

    def check_same_resolution(self, d1, d2):
        if d1.resolution != d2.resolution:
            raise ValueError(f"Incompatible resolutions: {d1.resolution} and {d2.resolution} ({d1} {d2})")

    def check_same_frequency(self, d1, d2):
        if d1.frequency != d2.frequency:
            raise ValueError(f"Incompatible frequencies: {d1.frequency} and {d2.frequency} ({d1} {d2})")

    def check_same_grid(self, d1, d2):
        if (d1.latitudes != d2.latitudes).any() or (d1.longitudes != d2.longitudes).any():
            raise ValueError(f"Incompatible grid ({d1} {d2})")

    def check_same_shape(self, d1, d2):
        if d1.shape[1:] != d2.shape[1:]:
            raise ValueError(f"Incompatible shapes: {d1.shape} and {d2.shape} ({d1} {d2})")

        if d1.variables != d2.variables:
            raise ValueError(f"Incompatible variables: {d1.variables} and {d2.variables} ({d1} {d2})")

    def check_same_sub_shapes(self, d1, d2, drop_axis):
        shape1 = d1.sub_shape(drop_axis)
        shape2 = d2.sub_shape(drop_axis)

        if shape1 != shape2:
            raise ValueError(f"Incompatible shapes: {d1.shape} and {d2.shape} ({d1} {d2})")

    def check_same_variables(self, d1, d2):
        if d1.variables != d2.variables:
            raise ValueError(f"Incompatible variables: {d1.variables} and {d2.variables} ({d1} {d2})")

    def check_same_lengths(self, d1, d2):
        if d1._len != d2._len:
            raise ValueError(f"Incompatible lengths: {d1._len} and {d2._len}")

    def check_same_dates(self, d1, d2):
        self.check_same_frequency(d1, d2)

        if d1.dates[0] != d2.dates[0]:
            raise ValueError(f"Incompatible start dates: {d1.dates[0]} and {d2.dates[0]} ({d1} {d2})")

        if d1.dates[-1] != d2.dates[-1]:
            raise ValueError(f"Incompatible end dates: {d1.dates[-1]} and {d2.dates[-1]} ({d1} {d2})")

    def check_compatibility(self, d1, d2):
        # These are the default checks
        # Derived classes should turn individual checks off if they are not needed
        self.check_same_resolution(d1, d2)
        self.check_same_frequency(d1, d2)
        self.check_same_grid(d1, d2)
        self.check_same_lengths(d1, d2)
        self.check_same_variables(d1, d2)
        self.check_same_dates(d1, d2)

    def provenance(self):
        return [d.provenance() for d in self.datasets]

    def __repr__(self):
        lst = ", ".join(repr(d) for d in self.datasets)
        return f"{self.__class__.__name__}({lst})"

    def metadata_specific(self, **kwargs):
        # We need to skip the forward superclass
        # TODO: revisit this
        return Dataset.metadata_specific(
            self,
            datasets=[d.metadata_specific() for d in self.datasets],
            **kwargs,
        )

    @cached_property
    def missing(self):
        offset = 0
        result = set()
        for d in self.datasets:
            result.update(offset + m for m in d.missing)
            offset += len(d)
        return result


class GivenAxis(Combined):
    """Given a given axis, combine the datasets along that axis."""

    def __init__(self, datasets, axis):
        self.axis = axis
        super().__init__(datasets)

        assert axis > 0 and axis < len(self.datasets[0].shape), (
            axis,
            self.datasets[0].shape,
        )

    def check_compatibility(self, d1, d2):
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=self.axis)

    @cached_property
    def shape(self):
        shapes = [d.shape for d in self.datasets]
        before = shapes[0][: self.axis]
        after = shapes[0][self.axis + 1 :]
        result = before + (sum(s[self.axis] for s in shapes),) + after
        assert False not in result, result
        return result

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        lengths = [d.shape[self.axis] for d in self.datasets]
        slices = length_to_slices(index[self.axis], lengths)
        result = [d[update_tuple(index, self.axis, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=self.axis)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def _get_slice(self, s):
        return np.stack([self[i] for i in range(*s.indices(self._len))])

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return np.concatenate([d[n] for d in self.datasets], axis=self.axis - 1)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import calendar
import datetime
import logging
import os
import re
from pathlib import PurePath

import numpy as np
import zarr

from .dataset import Dataset

LOG = logging.getLogger(__name__)

CONFIG = None

try:
    import tomllib  # Only available since 3.11
except ImportError:
    import tomli as tomllib


def add_named_dataset(name, path, **kwargs):
    load_config()
    if name in CONFIG["datasets"]["named"]:
        raise ValueError(f"Dataset {name} already exists")

    CONFIG["datasets"]["named"][name] = path


def add_dataset_path(path):
    load_config()

    if path not in CONFIG["datasets"]["path"]:
        CONFIG["datasets"]["path"].append(path)

    # save_config()


def load_config():
    global CONFIG
    if CONFIG is not None:
        return CONFIG

    conf = os.path.expanduser("~/.anemoi.toml")

    if os.path.exists(conf):

        with open(conf, "rb") as f:
            CONFIG = tomllib.load(f)
    else:
        CONFIG = {}

    CONFIG.setdefault("datasets", {})
    CONFIG["datasets"].setdefault("path", [])
    CONFIG["datasets"].setdefault("named", {})

    return CONFIG


def _frequency_to_hours(frequency):
    if isinstance(frequency, int):
        return frequency

    if isinstance(frequency, float):
        assert int(frequency) == frequency
        return int(frequency)

    m = re.match(r"(\d+)([dh])?", frequency)
    if m is None:
        raise ValueError("Invalid frequency: " + frequency)

    frequency = int(m.group(1))
    if m.group(2) == "h":
        return frequency

    if m.group(2) == "d":
        return frequency * 24

    raise NotImplementedError()


def _as_date(d, dates, last):
    if isinstance(d, np.datetime64):
        d = d.astype(datetime.datetime)

    if isinstance(d, datetime.datetime):
        if not d.minute == 0 and d.hour == 0 and d.second == 0:
            return np.datetime64(d)
        d = datetime.date(d.year, d.month, d.day)

    if isinstance(d, datetime.date):
        d = d.year * 10_000 + d.month * 100 + d.day

    try:
        d = int(d)
    except (ValueError, TypeError):
        pass

    if isinstance(d, int):
        if len(str(d)) == 4:
            year = d
            if last:
                return np.datetime64(f"{year:04}-12-31T23:59:59")
            else:
                return np.datetime64(f"{year:04}-01-01T00:00:00")

        if len(str(d)) == 6:
            year = d // 100
            month = d % 100
            if last:
                _, last_day = calendar.monthrange(year, month)
                return np.datetime64(f"{year:04}-{month:02}-{last_day:02}T23:59:59")
            else:
                return np.datetime64(f"{year:04}-{month:02}-01T00:00:00")

        if len(str(d)) == 8:
            year = d // 10000
            month = (d % 10000) // 100
            day = d % 100
            if last:
                return np.datetime64(f"{year:04}-{month:02}-{day:02}T23:59:59")
            else:
                return np.datetime64(f"{year:04}-{month:02}-{day:02}T00:00:00")

    if isinstance(d, str):

        if "-" in d and ":" in d:
            date, time = d.replace(" ", "T").split("T")
            year, month, day = date.split("-")
            hour, minute, second = time.split(":")
            return np.datetime64(f"{year:04}-{month:02}-{day:02}T{hour}:{minute}:{second}")

        if "-" in d:
            assert ":" not in d
            bits = d.split("-")
            if len(bits) == 1:
                return _as_date(int(bits[0]), dates, last)

            if len(bits) == 2:
                return _as_date(int(bits[0]) * 100 + int(bits[1]), dates, last)

            if len(bits) == 3:
                return _as_date(
                    int(bits[0]) * 10000 + int(bits[1]) * 100 + int(bits[2]),
                    dates,
                    last,
                )
        if ":" in d:
            assert len(d) == 5
            hour, minute = d.split(":")
            assert minute == "00"
            assert not last
            first = dates[0].astype(object)
            year = first.year
            month = first.month
            day = first.day

            return np.datetime64(f"{year:04}-{month:02}-{day:02}T{hour}:00:00")

    raise NotImplementedError(f"Unsupported date: {d} ({type(d)})")


def as_first_date(d, dates):
    return _as_date(d, dates, last=False)


def as_last_date(d, dates):
    return _as_date(d, dates, last=True)


def _concat_or_join(datasets, kwargs):

    if "adjust" in kwargs:
        raise ValueError("Cannot use 'adjust' without specifying 'concat' or 'join'")
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    # Study the dates
    ranges = [(d.dates[0].astype(object), d.dates[-1].astype(object)) for d in datasets]

    if len(set(ranges)) == 1:
        from .join import Join

        return Join(datasets)._overlay(), kwargs

    # Make sure the dates are disjoint
    for i in range(len(ranges)):
        r = ranges[i]
        for j in range(i + 1, len(ranges)):
            s = ranges[j]
            if r[0] <= s[0] <= r[1] or r[0] <= s[1] <= r[1]:
                raise ValueError(f"Overlapping dates: {r} and {s} ({datasets[i]} {datasets[j]})")

    # For now we should have the datasets in order with no gaps

    frequency = _frequency_to_hours(datasets[0].frequency)

    for i in range(len(ranges) - 1):
        r = ranges[i]
        s = ranges[i + 1]
        if r[1] + datetime.timedelta(hours=frequency) != s[0]:
            raise ValueError(
                "Datasets must be sorted by dates, with no gaps: " f"{r} and {s} ({datasets[i]} {datasets[i+1]})"
            )

    from .concat import Concat

    return Concat(datasets), kwargs


def _open(a):
    from .stores import Zarr
    from .stores import zarr_lookup

    if isinstance(a, Dataset):
        return a

    if isinstance(a, zarr.hierarchy.Group):
        return Zarr(a).mutate()

    if isinstance(a, str):
        return Zarr(zarr_lookup(a)).mutate()

    if isinstance(a, PurePath):
        return _open(str(a))

    if isinstance(a, dict):
        return _open_dataset(**a)

    if isinstance(a, (list, tuple)):
        return _open_dataset(*a)

    raise NotImplementedError("Unsupported argument: " + type(a))


def _auto_adjust(datasets, kwargs):

    if "adjust" not in kwargs:
        return datasets, kwargs

    adjust_list = kwargs.pop("adjust")
    if not isinstance(adjust_list, (tuple, list)):
        adjust_list = [adjust_list]

    ALIASES = {
        "all": ["select", "frequency", "start", "end"],
        "dates": ["start", "end", "frequency"],
        "variables": ["select"],
    }

    adjust_set = set()

    for a in adjust_list:
        adjust_set.update(ALIASES.get(a, [a]))

    extra = set(adjust_set) - set(ALIASES["all"])
    if extra:
        raise ValueError(f"Invalid adjust keys: {extra}")

    subset_kwargs = [{} for _ in datasets]

    if "select" in adjust_set:
        assert "select" not in kwargs, "Cannot use 'select' in adjust and kwargs"

        variables = None

        for d in datasets:
            if variables is None:
                variables = set(d.variables)
            else:
                variables &= set(d.variables)

        if len(variables) == 0:
            raise ValueError("No common variables")

        for i, d in enumerate(datasets):
            if set(d.variables) != variables:
                subset_kwargs[i]["select"] = sorted(variables)

    if "start" in adjust_set:
        assert "start" not in kwargs, "Cannot use 'start' in adjust and kwargs"
        start = max(d.dates[0] for d in datasets).astype(object)
        for i, d in enumerate(datasets):
            if start != d.dates[0]:
                subset_kwargs[i]["start"] = start

    if "end" in adjust_set:
        assert "end" not in kwargs, "Cannot use 'end' in adjust and kwargs"
        end = min(d.dates[-1] for d in datasets).astype(object)
        for i, d in enumerate(datasets):
            if end != d.dates[-1]:
                subset_kwargs[i]["end"] = end

    if "frequency" in adjust_set:
        assert "frequency" not in kwargs, "Cannot use 'frequency' in adjust and kwargs"
        frequency = max(d.frequency for d in datasets)
        for i, d in enumerate(datasets):
            if d.frequency != frequency:
                subset_kwargs[i]["frequency"] = frequency

    datasets = [d._subset(**subset_kwargs[i]) for i, d in enumerate(datasets)]

    return datasets, kwargs


def _open_dataset(*args, **kwargs):
    sets = []
    for a in args:
        sets.append(_open(a))

    if "zip" in kwargs:
        from .unchecked import zip_factory

        assert not sets, sets
        return zip_factory(args, kwargs)

    if "chain" in kwargs:
        from .unchecked import chain_factory

        assert not sets, sets
        return chain_factory(args, kwargs)

    if "join" in kwargs:
        from .join import join_factory

        assert not sets, sets
        return join_factory(args, kwargs)

    if "concat" in kwargs:
        from .concat import concat_factory

        assert not sets, sets
        return concat_factory(args, kwargs)

    if "ensemble" in kwargs:
        from .ensemble import ensemble_factory

        assert not sets, sets
        return ensemble_factory(args, kwargs)

    if "grids" in kwargs:
        from .grids import grids_factory

        assert not sets, sets
        return grids_factory(args, kwargs)

    if "cutout" in kwargs:
        from .grids import cutout_factory

        assert not sets, sets
        return cutout_factory(args, kwargs)

    for name in ("datasets", "dataset"):
        if name in kwargs:
            datasets = kwargs.pop(name)
            if not isinstance(datasets, (list, tuple)):
                datasets = [datasets]
            for a in datasets:
                sets.append(_open(a))

    assert len(sets) > 0, (args, kwargs)

    if len(sets) > 1:
        dataset, kwargs = _concat_or_join(sets, kwargs)
        return dataset._subset(**kwargs)

    return sets[0]._subset(**kwargs)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from functools import wraps

from .concat import ConcatMixin
from .debug import Node
from .forewards import Combined
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class check:

    def __init__(self, check):
        self.check = check

    def __call__(self, method):
        name = method.__name__
        check = self.check

        @wraps(method)
        def wrapper(obj):
            """
            This is a decorator that checks the compatibility of the datasets
            before calling the method. If the datasets are compatible, it
            will return the result of the method, otherwise it will raise an
            exception.
            """

            for d in obj.datasets[1:]:
                getattr(obj, check)(obj.datasets[0], d)

            return getattr(Combined, name).__get__(obj)

        return wrapper


class Unchecked(Combined):

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])

    def _subset(self, **kwargs):
        assert not kwargs
        return self

    def check_compatibility(self, d1, d2):
        pass

    ###########################################
    @property
    @check("check_same_dates")
    def dates(self):
        pass

    @property
    @check("check_same_resolution")
    def resolution(self):
        pass

    @property
    def field_shape(self):
        raise NotImplementedError()

    @property
    @check("check_same_frequency")
    def frequency(self):
        raise NotImplementedError()

    @property
    @check("check_same_grid")
    def latitudes(self):
        raise NotImplementedError()

    @property
    @check("check_same_grid")
    def longitudes(self):
        raise NotImplementedError()

    @property
    @check("check_same_variables")
    def name_to_index(self):
        raise NotImplementedError()

    @property
    @check("check_same_variables")
    def variables(self):
        raise NotImplementedError()

    @property
    @check("check_same_variables")
    def statistics(self):
        raise NotImplementedError()

    @property
    def shape(self):
        raise NotImplementedError()

    @property
    def dtype(self):
        raise NotImplementedError()

    @property
    def grids(self):
        raise NotImplementedError()


class Zip(Unchecked):

    def __len__(self):
        return min(len(d) for d in self.datasets)

    def __getitem__(self, n):
        return tuple(d[n] for d in self.datasets)

    @cached_property
    def missing(self):
        result = set()
        for d in self.datasets:
            result = result | d.missing
        return result


class Chain(ConcatMixin, Unchecked):
    """
    Same as Concat, but with no checks
    """

    def __len__(self):
        return sum(len(d) for d in self.datasets)

    def __getitem__(self, n):
        return tuple(d[n] for d in self.datasets)

    @property
    def dates(self):
        raise NotImplementedError()


def zip_factory(args, kwargs):

    zip = kwargs.pop("zip")
    assert len(args) == 0
    assert isinstance(zip, (list, tuple))

    datasets = [_open(e) for e in zip]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Zip(datasets)._subset(**kwargs)


def chain_factory(args, kwargs):

    chain = kwargs.pop("chain")
    assert len(args) == 0
    assert isinstance(chain, (list, tuple))

    datasets = [_open(e) for e in chain]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Chain(datasets)._subset(**kwargs)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from .debug import Node
from .debug import debug_indexing
from .forewards import Combined
from .forewards import GivenAxis
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Concat(Combined):
    def __len__(self):
        return sum(len(i) for i in self.datasets)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        # print(index, changes)
        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(index[0], lengths)
        # print("slies", slices)
        result = [d[update_tuple(index, 0, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=0)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        # TODO: optimize
        k = 0
        while n >= self.datasets[k]._len:
            n -= self.datasets[k]._len
            k += 1
        return self.datasets[k][n]

    @debug_indexing
    def _get_slice(self, s):
        result = []

        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(s, lengths)

        result = [d[i] for (d, i) in zip(self.datasets, slices) if i is not None]

        return np.concatenate(result)

    def check_compatibility(self, d1, d2):
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def check_same_lengths(self, d1, d2):
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1, d2):
        # Turned off because we are concatenating along the dates axis
        pass

    @property
    def dates(self):
        return np.concatenate([d.dates for d in self.datasets])

    @property
    def shape(self):
        return (len(self),) + self.datasets[0].shape[1:]

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])


class Grids(GivenAxis):
    def __init__(self, datasets, axis):
        super().__init__(datasets, axis)
        # Shape: (dates, variables, ensemble, 1d-values)
        assert len(datasets[0].shape) == 4, "Grids must be 1D for now"

    def check_same_grid(self, d1, d2):
        # We don't check the grid, because we want to be able to combine
        pass

    def check_same_resolution(self, d1, d2):
        # We don't check the resolution, because we want to be able to combine
        pass


class ConcatGrids(Grids):
    # TODO: select the statistics of the most global grid?
    @property
    def latitudes(self):
        return np.concatenate([d.latitudes for d in self.datasets])

    @property
    def longitudes(self):
        return np.concatenate([d.longitudes for d in self.datasets])

    @property
    def grids(self):
        result = []
        for d in self.datasets:
            result.extend(d.grids)
        return tuple(result)

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets], mode="concat")


class CutoutGrids(Grids):
    def __init__(self, datasets, axis):
        from anemoi.datasets.grids import cutout_mask

        super().__init__(datasets, axis)
        assert len(datasets) == 2, "CutoutGrids requires two datasets"
        assert axis == 3, "CutoutGrids requires axis=3"

        # We assume that the LAM is the first dataset, and the global is the second
        # Note: the second fields does not really need to be global

        self.lam, self.globe = datasets
        self.mask = cutout_mask(
            self.lam.latitudes,
            self.lam.longitudes,
            self.globe.latitudes,
            self.globe.longitudes,
            # plot="cutout",
        )
        assert len(self.mask) == self.globe.shape[3], (
            len(self.mask),
            self.globe.shape[3],
        )

    @cached_property
    def shape(self):
        shape = self.lam.shape
        # Number of non-zero masked values in the globe dataset
        nb_globe = np.count_nonzero(self.mask)
        return shape[:-1] + (shape[-1] + nb_globe,)

    def check_same_resolution(self, d1, d2):
        # Turned off because we are combining different resolutions
        pass

    @property
    def latitudes(self):
        return np.concatenate([self.lam.latitudes, self.globe.latitudes[self.mask]])

    @property
    def longitudes(self):
        return np.concatenate([self.lam.longitudes, self.globe.longitudes[self.mask]])

    def __getitem__(self, index):
        if isinstance(index, (int, slice)):
            index = (index, slice(None), slice(None), slice(None))
        return self._get_tuple(index)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        assert self.axis >= len(index) or index[self.axis] == slice(
            None
        ), f"No support for selecting a subset of the 1D values {index} ({self.tree()})"
        index, changes = index_to_slices(index, self.shape)

        # In case index_to_slices has changed the last slice
        index, _ = update_tuple(index, self.axis, slice(None))

        lam_data = self.lam[index]
        globe_data = self.globe[index]

        globe_data = globe_data[:, :, :, self.mask]

        result = np.concatenate([lam_data, globe_data], axis=self.axis)

        return apply_index_to_slices_changes(result, changes)

    @property
    def grids(self):
        for d in self.datasets:
            if len(d.grids) > 1:
                raise NotImplementedError("CutoutGrids does not support multi-grids datasets as inputs")
        shape = self.lam.shape
        return (shape[-1], self.shape[-1] - shape[-1])

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])


def grids_factory(args, kwargs):
    if "ensemble" in kwargs:
        raise NotImplementedError("Cannot use both 'ensemble' and 'grids'")

    grids = kwargs.pop("grids")
    axis = kwargs.pop("axis", 3)

    assert len(args) == 0
    assert isinstance(grids, (list, tuple))

    datasets = [_open(e) for e in grids]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return ConcatGrids(datasets, axis=axis)._subset(**kwargs)


def cutout_factory(args, kwargs):
    if "ensemble" in kwargs:
        raise NotImplementedError("Cannot use both 'ensemble' and 'cutout'")

    cutout = kwargs.pop("cutout")
    axis = kwargs.pop("axis", 3)

    assert len(args) == 0
    assert isinstance(cutout, (list, tuple))

    datasets = [_open(e) for e in cutout]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return CutoutGrids(datasets, axis=axis)._subset(**kwargs)



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from .misc import _open_dataset
from .misc import add_dataset_path
from .misc import add_named_dataset

LOG = logging.getLogger(__name__)

__all__ = [
    "open_dataset",
    "MissingDateError",
    "add_dataset_path",
    "add_named_dataset",
]


class MissingDateError(Exception):
    pass


def open_dataset(*args, **kwargs):
    ds = _open_dataset(*args, **kwargs)
    ds.arguments = {"args": args, "kwargs": kwargs}
    ds._check()
    return ds



#!/usr/bin/env python3
import json
import os

import zarr


def fix_order_by(order_by):
    if isinstance(order_by, list):
        return order_by

    assert isinstance(order_by, dict), order_by
    assert len(order_by) <= 3, order_by
    lst = []
    lst.append({"valid_datetime": order_by["valid_datetime"]})
    lst.append({"param_level": order_by["param_level"]})
    lst.append({"number": order_by["number"]})
    return lst


def fix_history(history):
    new = history
    new = [d for d in new if d.get("action") != "loading_data_start"]
    new = [d for d in new if d.get("action") != "loading_data_end"]
    return new


def fix_provenance(provenance):
    if "python" not in provenance:
        provenance["python"] = provenance["platform"]["python_version"]

    for q in (
        "args",
        "config_paths",
        "executable",
        "gpus",
        "platform",
        "python_path",
        "assets",
    ):
        if q in provenance:
            del provenance[q]

    for k, v in list(provenance["module_versions"].items()):
        if v.startswith("<"):
            del provenance["module_versions"][k]
        if v.startswith("/"):
            provenance["module_versions"][k] = os.path.join("...", os.path.basename(v))

    for k, v in list(provenance["git_versions"].items()):
        print(k, v)
        modified_files = v["git"].get("modified_files", [])
        untracked_files = v["git"].get("untracked_files", [])
        if not isinstance(modified_files, int):
            modified_files = len(modified_files)
        if not isinstance(untracked_files, int):
            untracked_files = len(untracked_files)
        provenance["git_versions"][k] = dict(
            git={
                "sha1": v["git"]["sha1"],
                "modified_files": modified_files,
                "untracked_files": untracked_files,
            }
        )

    print(json.dumps(provenance, indent=2))
    # assert False
    return provenance


def apply_patch(path, verbose=True, dry_run=False):
    print("====================")
    print(f"Patching {path}")
    print("====================")

    try:
        attrs = zarr.open(path, mode="r").attrs.asdict()
    except zarr.errors.PathNotFoundError as e:
        print(f"Failed to open {path}")
        print(e)
        exit(0)

    FIXES = {
        "history": fix_history,
        "provenance_load": fix_provenance,
        "provenance_statistics": fix_provenance,
        "order_by": fix_order_by,
    }
    REMOVE = ["_create_yaml_config"]

    before = json.dumps(attrs, sort_keys=True)

    fixed_attrs = {}
    for k, v in attrs.items():
        v = attrs[k]
        if k in REMOVE:
            print(f"✅ Remove {k}")
            continue

        if k not in FIXES:
            assert not k.startswith("provenance"), f"[{k}]"
            print(f"✅ Don't fix {k}")
            fixed_attrs[k] = v
            continue

        new_v = FIXES[k](v)
        if json.dumps(new_v, sort_keys=True) != json.dumps(v, sort_keys=True):
            print(f"✅ Fix {k}")
            if verbose:
                print(f"  Before : {k}= {v}")
                print(f"  After  : {k}= {new_v}")
        else:
            print(f"✅ Unchanged {k}")
        fixed_attrs[k] = new_v

    if dry_run:
        return
    z = zarr.open(path, mode="r+")

    for k in list(z.attrs.keys()):
        if k not in fixed_attrs:
            del z.attrs[k]
    for k, v in fixed_attrs.items():
        z.attrs[k] = v

    after = json.dumps(z.attrs.asdict(), sort_keys=True)
    if before != after:
        print("CHANGED")

    assert json.dumps(z.attrs.asdict(), sort_keys=True) == json.dumps(fixed_attrs, sort_keys=True)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .forewards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import make_slice_or_index_from_list_or_tuple
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class Subset(Forwards):
    """Select a subset of the dates."""

    def __init__(self, dataset, indices, reason):
        while isinstance(dataset, Subset):
            indices = [dataset.indices[i] for i in indices]
            reason = {**reason, **dataset.reason}
            dataset = dataset.dataset

        self.dataset = dataset
        self.indices = list(indices)
        self.reason = {k: v for k, v in reason.items() if v is not None}

        # Forward other properties to the super dataset
        super().__init__(dataset)

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        assert n >= 0, n
        n = self.indices[n]
        return self.dataset[n]

    @debug_indexing
    def _get_slice(self, s):
        # TODO: check if the indices can be simplified to a slice
        # the time checking maybe be longer than the time saved
        # using a slice
        indices = [self.indices[i] for i in range(*s.indices(self._len))]
        indices = make_slice_or_index_from_list_or_tuple(indices)
        if isinstance(indices, slice):
            return self.dataset[indices]
        return np.stack([self.dataset[i] for i in indices])

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, n):
        index, changes = index_to_slices(n, self.shape)
        # print('INDEX', index, changes)
        indices = [self.indices[i] for i in range(*index[0].indices(self._len))]
        indices = make_slice_or_index_from_list_or_tuple(indices)
        # print('INDICES', indices)
        index, _ = update_tuple(index, 0, indices)
        result = self.dataset[index]
        result = apply_index_to_slices_changes(result, changes)
        return result

    def __len__(self):
        return len(self.indices)

    @cached_property
    def shape(self):
        return (len(self),) + self.dataset.shape[1:]

    @cached_property
    def dates(self):
        return self.dataset.dates[self.indices]

    @cached_property
    def frequency(self):
        dates = self.dates
        delta = dates[1].astype(object) - dates[0].astype(object)
        return int(delta.total_seconds() / 3600)

    def source(self, index):
        return Source(self, index, self.forward.source(index))

    def __repr__(self):
        return f"Subset({self.dataset},{self.dates[0]}...{self.dates[-1]}/{self.frequency})"

    @cached_property
    def missing(self):
        return {self.indices[i] for i in self.dataset.missing if i in self.indices}

    def tree(self):
        return Node(self, [self.dataset.tree()], **self.reason)
